#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Boot Image V2.0, V2.1."""

import logging
import os
from datetime import datetime
from typing import Iterator, List, Optional

from spsdk import SPSDKError
from spsdk.apps.utils import get_key
from spsdk.crypto.loaders import load_certificate_as_bytes
from spsdk.crypto.signature_provider import SignatureProvider
from spsdk.exceptions import SPSDKUnsupportedOperation
from spsdk.utils.crypto import CertBlockV2, Counter, crypto_backend
from spsdk.utils.crypto.abstract import BaseClass
from spsdk.utils.crypto.backend_internal import internal_backend
from spsdk.utils.crypto.certificate import Certificate
from spsdk.utils.crypto.common import calc_cypher_block_count
from spsdk.utils.misc import find_first, load_text, write_file

from . import sb_21_helper as elf2sb_helper21
from . import sly_bd_parser as bd_parser
from .commands import CmdHeader
from .headers import ImageHeaderV2
from .sections import BootSectionV2, CertSectionV2

logger = logging.getLogger(__name__)


class SBV2xAdvancedParams:
    """The class holds advanced parameters for the SB file encryption.

    These parameters are used for the tests; for production, use can use default values (random keys + current time)
    """

    @staticmethod
    def _create_nonce() -> bytes:
        """Return random nonce."""
        nonce = bytearray(crypto_backend().random_bytes(16))
        # clear nonce bit at offsets 31 and 63
        nonce[9] &= 0x7F
        nonce[13] &= 0x7F
        return bytes(nonce)

    def __init__(
        self,
        dek: Optional[bytes] = None,
        mac: Optional[bytes] = None,
        nonce: Optional[bytes] = None,
        timestamp: Optional[datetime] = None,
    ):
        """Initialize SBV2xAdvancedParams.

        :param dek: DEK key
        :param mac: MAC key
        :param nonce: nonce
        :param timestamp: fixed timestamp for the header; use None to use current date/time
        :raises SPSDKError: Invalid dek or mac
        :raises SPSDKError: Invalid length of nonce
        """
        self._dek: bytes = dek if dek else crypto_backend().random_bytes(32)
        self._mac: bytes = mac if mac else crypto_backend().random_bytes(32)
        self._nonce: bytes = nonce if nonce else SBV2xAdvancedParams._create_nonce()
        if timestamp is None:
            timestamp = datetime.now()
        self._timestamp = datetime.fromtimestamp(int(timestamp.timestamp()))
        if len(self._dek) != 32 and len(self._mac) != 32:
            raise SPSDKError("Invalid dek or mac")
        if len(self._nonce) != 16:
            raise SPSDKError("Invalid length of nonce")

    @property
    def dek(self) -> bytes:
        """Return DEK key."""
        return self._dek

    @property
    def mac(self) -> bytes:
        """Return MAC key."""
        return self._mac

    @property
    def nonce(self) -> bytes:
        """Return NONCE."""
        return self._nonce

    @property
    def timestamp(self) -> datetime:
        """Return timestamp."""
        return self._timestamp


########################################################################################################################
# Secure Boot Image Class (Version 2.0)
########################################################################################################################
class BootImageV20(BaseClass):
    """Boot Image V2.0 class."""

    # Image specific data
    # size of the MAC key
    HEADER_MAC_SIZE = 32
    # AES encrypted DEK and MAC, including padding
    DEK_MAC_SIZE = 32 + 32 + 16

    KEY_BLOB_SIZE = 80

    def __init__(
        self,
        signed: bool,
        kek: bytes,
        *sections: BootSectionV2,
        product_version: str = "1.0.0",
        component_version: str = "1.0.0",
        build_number: int = 0,
        advanced_params: SBV2xAdvancedParams = SBV2xAdvancedParams(),
    ) -> None:
        """Initialize Secure Boot Image V2.0.

        :param signed: True if image is signed, False otherwise
        :param kek: key for wrapping DEK and MAC keys
        :param product_version: The product version (default: 1.0.0)
        :param component_version: The component version (default: 1.0.0)
        :param build_number: The build number value (default: 0)
        :param advanced_params: Advanced parameters for encryption of the SB file, use for tests only
        :param sections: Boot sections
        :raises SPSDKError: Invalid dek or mac
        """
        self._kek = kek
        # Set Flags value
        self._signed = signed
        self._signature_provider: Optional[SignatureProvider] = None
        flags = 0x08 if self.signed else 0x04
        # Set private attributes
        self._dek: bytes = advanced_params.dek
        self._mac: bytes = advanced_params.mac
        if (
            len(self._dek) != self.HEADER_MAC_SIZE and len(self._mac) != self.HEADER_MAC_SIZE
        ):  # pragma: no cover # condition checked in SBV2xAdvancedParams constructor
            raise SPSDKError("Invalid dek or mac")
        self._header = ImageHeaderV2(
            version="2.0",
            product_version=product_version,
            component_version=component_version,
            build_number=build_number,
            flags=flags,
            nonce=advanced_params.nonce,
            timestamp=advanced_params.timestamp,
        )
        self._cert_section: Optional[CertSectionV2] = None
        self._boot_sections: List[BootSectionV2] = []
        # Generate nonce
        if self._header.nonce is None:
            nonce = bytearray(crypto_backend().random_bytes(16))
            # clear nonce bit at offsets 31 and 63
            nonce[9] &= 0x7F
            nonce[13] &= 0x7F
            self._header.nonce = bytes(nonce)
        # Sections
        for section in sections:
            self.add_boot_section(section)

    @property
    def header(self) -> ImageHeaderV2:
        """Return image header."""
        return self._header

    @property
    def dek(self) -> bytes:
        """Data encryption key."""
        return self._dek

    @property
    def mac(self) -> bytes:
        """Message authentication code."""
        return self._mac

    @property
    def kek(self) -> bytes:
        """Return key for wrapping DEK and MAC keys."""
        return self._kek

    @property
    def signature_provider(self) -> Optional[SignatureProvider]:
        """Signature provider instance to sign the final image.

        None if not assigned yet.
        """
        return self._signature_provider

    @signature_provider.setter
    def signature_provider(self, value: SignatureProvider) -> None:
        """Signature provider setter.

        :param value: Signature provider to sign final image
        """
        self._signature_provider = value

    @property
    def signed(self) -> bool:
        """Check whether sb is signed + encrypted or only encrypted."""
        return self._signed

    @property
    def cert_block(self) -> Optional[CertBlockV2]:
        """Return certificate block; None if SB file not signed or block not assigned yet."""
        cert_sect = self._cert_section
        if cert_sect is None:
            return None

        return cert_sect.cert_block

    @cert_block.setter
    def cert_block(self, value: Optional[CertBlockV2]) -> None:
        """Setter.

        :param value: block to be assigned; None to remove previously assigned block
        :raises SPSDKError: When certificate block is used when SB file is not signed
        """
        if value is not None:
            if not self.signed:
                raise SPSDKError("Certificate block cannot be used unless SB file is signed")
        self._cert_section = CertSectionV2(value) if value else None

    @property
    def cert_header_size(self) -> int:
        """Return image raw size (not aligned) for certificate header."""
        size = ImageHeaderV2.SIZE + self.HEADER_MAC_SIZE + self.KEY_BLOB_SIZE
        for boot_section in self._boot_sections:
            size += boot_section.raw_size
        return size

    @property
    def raw_size_without_signature(self) -> int:
        """Return image raw size without signature, used to calculate image blocks."""
        # Header, HMAC and KeyBlob
        size = ImageHeaderV2.SIZE + self.HEADER_MAC_SIZE + self.KEY_BLOB_SIZE
        # Certificates Section
        if self.signed:
            size += self.DEK_MAC_SIZE
            cert_block = self.cert_block
            if not cert_block:
                raise SPSDKError("Certification block not present")
            size += cert_block.raw_size
        # Boot Sections
        for boot_section in self._boot_sections:
            size += boot_section.raw_size
        return size

    @property
    def raw_size(self) -> int:
        """Return image raw size."""
        size = self.raw_size_without_signature

        if self.signed:
            cert_block = self.cert_block
            if not cert_block:  # pragma: no cover # already checked in raw_size_without_signature
                raise SPSDKError("Certificate block not present")
            size += cert_block.signature_size

        return size

    def __len__(self) -> int:
        return len(self._boot_sections)

    def __getitem__(self, key: int) -> BootSectionV2:
        return self._boot_sections[key]

    def __setitem__(self, key: int, value: BootSectionV2) -> None:
        self._boot_sections[key] = value

    def __iter__(self) -> Iterator[BootSectionV2]:
        return self._boot_sections.__iter__()

    def update(self) -> None:
        """Update boot image."""
        if self._boot_sections:
            self._header.first_boot_section_id = self._boot_sections[0].uid
            # calculate first boot tag block
            data_size = self._header.SIZE + self.HEADER_MAC_SIZE + self.KEY_BLOB_SIZE
            if self._cert_section is not None:
                data_size += self._cert_section.raw_size
            self._header.first_boot_tag_block = calc_cypher_block_count(data_size)
        # ...
        self._header.flags = 0x08 if self.signed else 0x04
        self._header.image_blocks = calc_cypher_block_count(self.raw_size_without_signature)
        self._header.header_blocks = calc_cypher_block_count(self._header.SIZE)
        self._header.max_section_mac_count = 0
        if self.signed:
            self._header.offset_to_certificate_block = (
                self._header.SIZE + self.HEADER_MAC_SIZE + self.KEY_BLOB_SIZE
            )
            self._header.offset_to_certificate_block += CmdHeader.SIZE + CertSectionV2.HMAC_SIZE * 2
            self._header.max_section_mac_count = 1
        for boot_sect in self._boot_sections:
            boot_sect.is_last = True  # this is unified with elftosb
            self._header.max_section_mac_count += boot_sect.hmac_count
        # Update certificates block header
        cert_blk = self.cert_block
        if cert_blk is not None:
            cert_blk.header.build_number = self._header.build_number
            cert_blk.header.image_length = self.cert_header_size

    def info(self) -> str:
        """Return text description of the instance."""
        self.update()
        nfo = "\n"
        nfo += ":::::::::::::::::::::::::::::::::: IMAGE HEADER ::::::::::::::::::::::::::::::::::::::\n"
        nfo += self._header.info()
        if self._cert_section is not None:
            nfo += "::::::::::::::::::::::::::::::: CERTIFICATES BLOCK ::::::::::::::::::::::::::::::::::::\n"
            nfo += self._cert_section.info()
        nfo += "::::::::::::::::::::::::::::::::::: BOOT SECTIONS ::::::::::::::::::::::::::::::::::::\n"
        for index, section in enumerate(self._boot_sections):
            nfo += f"[ SECTION: {index} | UID: 0x{section.uid:08X} ]\n"
            nfo += section.info()
        return nfo

    def add_boot_section(self, section: BootSectionV2) -> None:
        """Add new Boot section into image.

        :param section: Boot section
        :raises SPSDKError: Raised when section is not instance of BootSectionV2 class
        :raises SPSDKError: Raised when boot section has duplicate UID
        """
        if not isinstance(section, BootSectionV2):
            raise SPSDKError("Section is not instance of BootSectionV2 class")
        duplicate_uid = find_first(self._boot_sections, lambda bs: bs.uid == section.uid)
        if duplicate_uid is not None:
            raise SPSDKError(f"Boot section with duplicate UID: {str(section.uid)}")
        self._boot_sections.append(section)

    def export(self, padding: Optional[bytes] = None) -> bytes:
        """Serialize image object.

        :param padding: header padding (8 bytes) for testing purpose; None to use random values (recommended)
        :return: exported bytes
        :raises SPSDKError: Raised when there are no boot sections or is not signed or private keys are missing
        :raises SPSDKError: Raised when there is invalid dek or mac
        :raises SPSDKError: Raised when certificate data is not present
        :raises SPSDKError: Raised when there is invalid certificate block
        :raises SPSDKError: Raised when there is invalid length of exported data
        """
        if len(self.dek) != 32 or len(self.mac) != 32:
            raise SPSDKError("Invalid dek or mac")
        # validate params
        if not self._boot_sections:
            raise SPSDKError("No boot section")
        if self.signed and (self._cert_section is None):
            raise SPSDKError("Certificate section is required for signed images")
        # update internals
        self.update()
        # Add Image Header data
        data = self._header.export(padding=padding)
        # Add Image Header HMAC data
        data += crypto_backend().hmac(self.mac, data)
        # Add DEK and MAC keys
        data += crypto_backend().aes_key_wrap(self.kek, self.dek + self.mac)
        # Add Padding
        data += padding if padding else crypto_backend().random_bytes(8)
        # Add Certificates data
        if not self._header.nonce:
            raise SPSDKError("There is no nonce in the header")
        counter = Counter(self._header.nonce)
        counter.increment(calc_cypher_block_count(len(data)))
        if self._cert_section is not None:
            cert_sect_bin = self._cert_section.export(dek=self.dek, mac=self.mac, counter=counter)
            counter.increment(calc_cypher_block_count(len(cert_sect_bin)))
            data += cert_sect_bin
        # Add Boot Sections data
        for sect in self._boot_sections:
            data += sect.export(dek=self.dek, mac=self.mac, counter=counter)
        # Add Signature data
        if self.signed:
            if self.signature_provider is None:
                raise SPSDKError("Signature provider is not assigned, cannot sign the image.")
            if self.cert_block is None:
                raise SPSDKError("Certificate block is not assigned.")

            public_key = self.cert_block.certificates[-1].public_key
            try:
                result = self.signature_provider.verify_public_key(public_key.dump())
                if not result:
                    raise SPSDKError(
                        "Signature verification failed, public key does not match to certificate"
                    )
                logger.debug("The verification of private key pair integrity has been successful.")
            except SPSDKUnsupportedOperation:
                logger.warning(
                    "Signature provider could not verify the integrity of private key pair."
                )

            data += self.signature_provider.sign(data)

        if len(data) != self.raw_size:
            raise SPSDKError("Invalid length of exported data")
        return data

    # pylint: disable=too-many-locals
    @classmethod
    def parse(cls, data: bytes, offset: int = 0, kek: bytes = bytes()) -> "BootImageV20":
        """Parse image from bytes.

        :param data: Raw data of parsed image
        :param offset: The offset of input data
        :param kek: The Key for unwrapping DEK and MAC keys (required)
        :return: parsed image object
        :raises SPSDKError: raised when header is in wrong format
        :raises SPSDKError: raised when there is invalid header version
        :raises SPSDKError: raised when signature is incorrect
        :raises SPSDKError: Raised when kek is empty
        :raises SPSDKError: raised when header's nonce is not present
        """
        if not kek:
            raise SPSDKError("kek cannot be empty")
        index = offset
        header_raw_data = data[index : index + ImageHeaderV2.SIZE]
        index += ImageHeaderV2.SIZE
        header_mac_data = data[index : index + cls.HEADER_MAC_SIZE]
        index += cls.HEADER_MAC_SIZE
        key_blob = data[index : index + cls.KEY_BLOB_SIZE]
        index += cls.KEY_BLOB_SIZE
        key_blob_unwrap = crypto_backend().aes_key_unwrap(kek, key_blob[:-8])
        dek = key_blob_unwrap[:32]
        mac = key_blob_unwrap[32:]
        header_mac_data_calc = crypto_backend().hmac(mac, header_raw_data)
        if header_mac_data != header_mac_data_calc:
            raise SPSDKError("Invalid header MAC data")
        # Parse Header
        header = ImageHeaderV2.parse(header_raw_data)
        if header.version != "2.0":
            raise SPSDKError(f"Invalid Header Version: {header.version} instead 2.0")
        image_size = header.image_blocks * 16
        # Initialize counter
        if not header.nonce:
            raise SPSDKError("Header's nonce not present")
        counter = Counter(header.nonce)
        counter.increment(calc_cypher_block_count(index - offset))
        # ...
        signed = header.flags == 0x08
        adv_params = SBV2xAdvancedParams(
            dek=dek, mac=mac, nonce=header.nonce, timestamp=header.timestamp
        )
        obj = cls(
            signed,
            kek=kek,
            product_version=str(header.product_version),
            component_version=str(header.component_version),
            build_number=header.build_number,
            advanced_params=adv_params,
        )
        # Parse Certificate section
        if header.flags == 0x08:
            cert_sect = CertSectionV2.parse(data, index, dek=dek, mac=mac, counter=counter)
            obj._cert_section = cert_sect
            index += cert_sect.raw_size
            # Check Signature
            if not cert_sect.cert_block.verify_data(
                data[offset + image_size :], data[offset : offset + image_size]
            ):
                raise SPSDKError("Parsing Certification section failed")
        # Parse Boot Sections
        while index < (image_size + offset):
            boot_section = BootSectionV2.parse(data, index, dek=dek, mac=mac, counter=counter)
            obj.add_boot_section(boot_section)
            index += boot_section.raw_size
        return obj


########################################################################################################################
# Secure Boot Image Class (Version 2.1)
########################################################################################################################
class BootImageV21(BaseClass):
    """Boot Image V2.1 class."""

    # Image specific data
    HEADER_MAC_SIZE = 32
    KEY_BLOB_SIZE = 80
    SHA_256_SIZE = 32

    # defines
    FLAGS_SHA_PRESENT_BIT = 0x8000  # image contains SHA-256
    FLAGS_ENCRYPTED_SIGNED_BIT = 0x0008  # image is signed and encrypted

    def __init__(
        self,
        kek: bytes,
        *sections: BootSectionV2,
        product_version: str = "1.0.0",
        component_version: str = "1.0.0",
        build_number: int = 0,
        advanced_params: SBV2xAdvancedParams = SBV2xAdvancedParams(),
        flags: int = FLAGS_SHA_PRESENT_BIT | FLAGS_ENCRYPTED_SIGNED_BIT,
    ) -> None:
        """Initialize Secure Boot Image V2.1.

        :param kek: key to wrap DEC and MAC keys

        :param product_version: The product version (default: 1.0.0)
        :param component_version: The component version (default: 1.0.0)
        :param build_number: The build number value (default: 0)

        :param advanced_params: optional advanced parameters for encryption; it is recommended to use default value
        :param flags: see flags defined in class.
        :param sections: Boot sections
        """
        self._kek = kek
        self._signature_provider: Optional[
            SignatureProvider
        ] = None  # this should be assigned for export, not needed for parsing
        self._dek = advanced_params.dek
        self._mac = advanced_params.mac
        self._header = ImageHeaderV2(
            version="2.1",
            product_version=product_version,
            component_version=component_version,
            build_number=build_number,
            flags=flags,
            nonce=advanced_params.nonce,
            timestamp=advanced_params.timestamp,
        )
        self._cert_block: Optional[CertBlockV2] = None
        self.boot_sections: List[BootSectionV2] = []
        # ...
        for section in sections:
            self.add_boot_section(section)

    @property
    def header(self) -> ImageHeaderV2:
        """Return image header."""
        return self._header

    @property
    def dek(self) -> bytes:
        """Data encryption key."""
        return self._dek

    @property
    def mac(self) -> bytes:
        """Message authentication code."""
        return self._mac

    @property
    def kek(self) -> bytes:
        """Return key to wrap DEC and MAC keys."""
        return self._kek

    @property
    def signature_provider(self) -> Optional[SignatureProvider]:
        """Signature provider instance to sign the final image.

        None if not assigned yet.
        """
        return self._signature_provider

    @signature_provider.setter
    def signature_provider(self, value: SignatureProvider) -> None:
        """Signature provider setter.

        :param value: Signature provider to sign final image
        """
        self._signature_provider = value

    @property
    def cert_block(self) -> Optional[CertBlockV2]:
        """Return certificate block; None if SB file not signed or block not assigned yet."""
        return self._cert_block

    @cert_block.setter
    def cert_block(self, value: CertBlockV2) -> None:
        """Setter.

        :param value: block to be assigned; None to remove previously assigned block
        """
        assert isinstance(value, CertBlockV2)
        self._cert_block = value
        self._cert_block.alignment = 16

    @property
    def signed(self) -> bool:
        """Return flag whether SB file is signed."""
        return True  # SB2.1 is always signed

    @property
    def cert_header_size(self) -> int:
        """Return image raw size (not aligned) for certificate header."""
        size = ImageHeaderV2.SIZE + self.HEADER_MAC_SIZE
        size += self.KEY_BLOB_SIZE
        # Certificates Section
        cert_blk = self.cert_block
        if cert_blk:
            size += cert_blk.raw_size
        return size

    @property
    def raw_size(self) -> int:
        """Return image raw size (not aligned)."""
        # Header, HMAC and KeyBlob
        size = ImageHeaderV2.SIZE + self.HEADER_MAC_SIZE
        size += self.KEY_BLOB_SIZE
        # Certificates Section
        cert_blk = self.cert_block
        if cert_blk:
            size += cert_blk.raw_size
            if not self.signed:  # pragma: no cover # SB2.1 is always signed
                raise SPSDKError("Certificate block is not signed")
            size += cert_blk.signature_size
        # Boot Sections
        for boot_section in self.boot_sections:
            size += boot_section.raw_size
        return size

    def __len__(self) -> int:
        return len(self.boot_sections)

    def __getitem__(self, key: int) -> BootSectionV2:
        return self.boot_sections[key]

    def __setitem__(self, key: int, value: BootSectionV2) -> None:
        self.boot_sections[key] = value

    def __iter__(self) -> Iterator[BootSectionV2]:
        return self.boot_sections.__iter__()

    def update(self) -> None:
        """Update BootImageV21."""
        if self.boot_sections:
            self._header.first_boot_section_id = self.boot_sections[0].uid
            # calculate first boot tag block
            data_size = self._header.SIZE + self.HEADER_MAC_SIZE + self.KEY_BLOB_SIZE
            cert_blk = self.cert_block
            if cert_blk is not None:
                data_size += cert_blk.raw_size
                if not self.signed:  # pragma: no cover # SB2.1 is always signed
                    raise SPSDKError("Certificate block is not signed")
                data_size += cert_blk.signature_size
            self._header.first_boot_tag_block = calc_cypher_block_count(data_size)
        # ...
        self._header.image_blocks = calc_cypher_block_count(self.raw_size)
        self._header.header_blocks = calc_cypher_block_count(self._header.SIZE)
        self._header.offset_to_certificate_block = (
            self._header.SIZE + self.HEADER_MAC_SIZE + self.KEY_BLOB_SIZE
        )
        # Get HMAC count
        self._header.max_section_mac_count = 0
        for boot_sect in self.boot_sections:
            boot_sect.is_last = True  # unified with elftosb
            self._header.max_section_mac_count += boot_sect.hmac_count
        # Update certificates block header
        cert_clk = self.cert_block
        if cert_clk is not None:
            cert_clk.header.build_number = self._header.build_number
            cert_clk.header.image_length = self.cert_header_size

    def info(self) -> str:
        """Return text description of the instance."""
        self.update()
        nfo = "\n"
        nfo += ":::::::::::::::::::::::::::::::::: IMAGE HEADER ::::::::::::::::::::::::::::::::::::::\n"
        nfo += self._header.info()
        if self.cert_block is not None:
            nfo += "::::::::::::::::::::::::::::::: CERTIFICATES BLOCK ::::::::::::::::::::::::::::::::::::\n"
            nfo += self.cert_block.info()
        nfo += "::::::::::::::::::::::::::::::::::: BOOT SECTIONS ::::::::::::::::::::::::::::::::::::\n"
        for index, section in enumerate(self.boot_sections):
            nfo += f"[ SECTION: {index} | UID: 0x{section.uid:08X} ]\n"
            nfo += section.info()
        return nfo

    def add_boot_section(self, section: BootSectionV2) -> None:
        """Add new Boot section into image.

        :param section: Boot section to be added
        :raises SPSDKError: Raised when section is not instance of BootSectionV2 class
        """
        if not isinstance(section, BootSectionV2):
            raise SPSDKError("Section is not instance of BootSectionV2 class")
        self.boot_sections.append(section)

    # pylint: disable=too-many-locals
    def export(
        self, padding: Optional[bytes] = None, dbg_info: Optional[List[str]] = None
    ) -> bytes:
        """Serialize image object.

        :param padding: header padding (8 bytes) for testing purpose; None to use random values (recommended)
        :param dbg_info: optional list, where debug info is exported in text form
        :return: exported bytes
        :raises SPSDKError: Raised when there is no boot section to be added
        :raises SPSDKError: Raised when certificate is not assigned
        :raises SPSDKError: Raised when private key is not assigned
        :raises SPSDKError: Raised when private header's nonce is invalid
        :raises SPSDKError: Raised when private key does not match certificate
        :raises SPSDKError: Raised when there is no debug info
        """
        # validate params
        if not self.boot_sections:
            raise SPSDKError("At least one Boot Section must be added")
        if self.cert_block is None:
            raise SPSDKError("Certificate is not assigned")
        if self.signature_provider is None:
            raise SPSDKError("Signature provider is not assigned, cannot sign the image")
        # Update internals
        if dbg_info is not None:
            dbg_info.append("[sb_file]")
        bs_dbg_info: Optional[List[str]] = [] if dbg_info else None
        self.update()
        # Export Boot Sections
        bs_data = bytes()
        bs_offset = (
            ImageHeaderV2.SIZE
            + self.HEADER_MAC_SIZE
            + self.KEY_BLOB_SIZE
            + self.cert_block.raw_size
            + self.cert_block.signature_size
        )
        if self.header.flags & self.FLAGS_SHA_PRESENT_BIT:
            bs_offset += self.SHA_256_SIZE

        if not self._header.nonce:
            raise SPSDKError("Invalid header's nonce")
        counter = Counter(self._header.nonce, calc_cypher_block_count(bs_offset))
        for sect in self.boot_sections:
            bs_data += sect.export(
                dek=self.dek, mac=self.mac, counter=counter, dbg_info=bs_dbg_info
            )
        # Export Header
        signed_data = self._header.export(padding=padding)
        if dbg_info:
            dbg_info.append("[header]")
            dbg_info.append(signed_data.hex())
        #  Add HMAC data
        first_bs_hmac_count = self.boot_sections[0].hmac_count
        hmac_data = bs_data[CmdHeader.SIZE : CmdHeader.SIZE + (first_bs_hmac_count * 32) + 32]
        hmac = crypto_backend().hmac(self.mac, hmac_data)
        signed_data += hmac
        if dbg_info:
            dbg_info.append("[hmac]")
            dbg_info.append(hmac.hex())
        # Add KeyBlob data
        key_blob = crypto_backend().aes_key_wrap(self.kek, self.dek + self.mac)
        key_blob += b"\00" * (self.KEY_BLOB_SIZE - len(key_blob))
        signed_data += key_blob
        if dbg_info:
            dbg_info.append("[key_blob]")
            dbg_info.append(key_blob.hex())
        # Add Certificates data
        signed_data += self.cert_block.export()
        if dbg_info:
            dbg_info.append("[cert_block]")
            dbg_info.append(self.cert_block.export().hex())
        # Add SHA-256 of Bootable sections if requested
        if self.header.flags & self.FLAGS_SHA_PRESENT_BIT:
            signed_data += internal_backend.hash(bs_data)

        # Add Signature data
        public_key = self.cert_block.certificates[-1].public_key
        try:
            result = self.signature_provider.verify_public_key(public_key.dump())
            if not result:
                raise SPSDKError(
                    "Signature verification failed, public key does not match to certificate"
                )
            logger.debug("The verification of private key pair integrity has been successful.")
        except SPSDKUnsupportedOperation:
            logger.warning("Signature provider could not verify the integrity of private key pair.")

        signature = self.signature_provider.sign(signed_data)
        if dbg_info:
            dbg_info.append("[signature]")
            dbg_info.append(signature.hex())
            dbg_info.append("[boot_sections]")
            if not bs_dbg_info:
                raise SPSDKError("No debug information")
            dbg_info.extend(bs_dbg_info)
        return signed_data + signature + bs_data

    # pylint: disable=too-many-locals
    @classmethod
    def parse(
        cls,
        data: bytes,
        offset: int = 0,
        kek: bytes = bytes(),
        plain_sections: bool = False,
    ) -> "BootImageV21":
        """Parse image from bytes.

        :param data: Raw data of parsed image
        :param offset: The offset of input data
        :param kek: The Key for unwrapping DEK and MAC keys (required)
        :param plain_sections: Sections are not encrypted; this is used only for debugging,
            not supported by ROM code
        :return: BootImageV21 parsed object
        :raises SPSDKError: raised when header is in incorrect format
        :raises SPSDKError: raised when signature is incorrect
        :raises SPSDKError: Raised when kek is empty
        :raises SPSDKError: raised when header's nonce not present"
        """
        if not kek:
            raise SPSDKError("kek cannot be empty")
        index = offset
        header_raw_data = data[index : index + ImageHeaderV2.SIZE]
        index += ImageHeaderV2.SIZE
        # Not used right now: hmac_data = data[index: index + cls.HEADER_MAC_SIZE]
        index += cls.HEADER_MAC_SIZE
        key_blob = data[index : index + cls.KEY_BLOB_SIZE]
        index += cls.KEY_BLOB_SIZE
        key_blob_unwrap = crypto_backend().aes_key_unwrap(kek, key_blob[:-8])
        dek = key_blob_unwrap[:32]
        mac = key_blob_unwrap[32:]
        # Parse Header
        header = ImageHeaderV2.parse(header_raw_data)
        if header.offset_to_certificate_block != (index - offset):
            raise SPSDKError("Invalid offset")
        # Parse Certificate Block
        cert_block = CertBlockV2.parse(data, index)
        index += cert_block.raw_size

        # Verify Signature
        signature_index = index
        # The image may contain SHA, in such a case the signature is placed
        # after SHA. Thus we must shift the index by SHA size.
        if header.flags & BootImageV21.FLAGS_SHA_PRESENT_BIT:
            signature_index += BootImageV21.SHA_256_SIZE
        result = cert_block.verify_data(
            data[signature_index : signature_index + cert_block.signature_size],
            data[offset:signature_index],
        )

        if not result:
            raise SPSDKError("Verification failed")
        # Check flags, if 0x8000 bit is set, the SB file contains SHA-256 between
        # certificate and signature.
        if header.flags & BootImageV21.FLAGS_SHA_PRESENT_BIT:
            bootable_section_sha256 = data[index : index + BootImageV21.SHA_256_SIZE]
            index += BootImageV21.SHA_256_SIZE
        index += cert_block.signature_size
        # Check first Boot Section HMAC
        # Not implemented yet
        # hmac_data_calc = crypto_backend().hmac(mac, data[index + CmdHeader.SIZE: index + CmdHeader.SIZE + ((2) * 32)])
        # if hmac_data != hmac_data_calc:
        #    raise SPSDKError("HMAC failed")
        if not header.nonce:
            raise SPSDKError("Header's nonce not present")
        counter = Counter(header.nonce)
        counter.increment(calc_cypher_block_count(index - offset))
        boot_section = BootSectionV2.parse(
            data, index, dek=dek, mac=mac, counter=counter, plain_sect=plain_sections
        )
        if header.flags & BootImageV21.FLAGS_SHA_PRESENT_BIT:
            computed_bootable_section_sha256 = internal_backend.hash(
                data[index:], algorithm="sha256"
            )

            if bootable_section_sha256 != computed_bootable_section_sha256:
                raise SPSDKError(
                    desc=(
                        "Error: invalid Bootable section SHA."
                        f"Expected {bootable_section_sha256.decode('utf-8')},"
                        f"got {computed_bootable_section_sha256.decode('utf-8')}"
                    )
                )
        adv_params = SBV2xAdvancedParams(
            dek=dek, mac=mac, nonce=header.nonce, timestamp=header.timestamp
        )
        obj = cls(
            kek=kek,
            product_version=str(header.product_version),
            component_version=str(header.component_version),
            build_number=header.build_number,
            advanced_params=adv_params,
        )
        obj.cert_block = cert_block
        obj.add_boot_section(boot_section)
        return obj


def generate_SB21(  # pylint: disable=invalid-name
    bd_file_path: str,
    key_file_path: str,
    signature_provider: SignatureProvider,
    signing_certificate_file_paths: List[str],
    root_key_certificate_paths: List[str],
    hoh_out_path: str,
    external_files: List[str],
) -> bytes:
    """Generate SecureBinary image from BD command file.

    :param bd_file_path: path to BD file.
    :param key_file_path: path to key file.
    :param signature_provider: Signature provider to sign final image
    :param signing_certificate_file_paths: signing certificate chain.
    :param root_key_certificate_paths: paths to root key certificate(s) for
        verifying other certificates. Only 4 root key certificates are allowed,
        others are ignored. One of the certificates must match the first certificate
        passed in signing_certificate_file_paths.
    :param hoh_out_path: output path to hash of hashes of root keys. If set to
        None, 'hash.bin' is created under working directory.
    :param external_files: external files referenced from BD file.

    :raises SPSDKError: If incorrect bd file is provided
    """
    # Create lexer and parser, load the BD file content and parse it for
    # further execution - the parsed BD file is a dictionary in JSON format
    bd_file_content = load_text(bd_file_path)

    parser = bd_parser.BDParser()

    parsed_bd_file = parser.parse(text=bd_file_content, extern=external_files)
    if parsed_bd_file is None:
        raise SPSDKError("Invalid bd file, secure binary file generation terminated")

    # The dictionary contains following content:
    # {
    #   options: {
    #       opt1: value,...
    #   },
    #   sections: [
    #       {section_id: value, options: {}, commands: {}},
    #       {section_id: value, options: {}, commands: {}}
    #   ]
    # }
    # TODO check, that section_ids differ in sections???

    # we need to encrypt and sign the image, let's check, whether we have
    # everything we need
    # It appears, that flags option in BD file are irrelevant for 2.1 secure
    # binary images regarding encryption/signing - SB 2.1 must be encrypted
    # and signed.
    # However, bit 15 represents, whether the final SB 2.1 must include a
    # SHA-256 of the bootable section.
    flags = parsed_bd_file["options"].get(
        "flags", BootImageV21.FLAGS_SHA_PRESENT_BIT | BootImageV21.FLAGS_ENCRYPTED_SIGNED_BIT
    )
    if (
        signature_provider is None
        or signing_certificate_file_paths is None
        or root_key_certificate_paths is None
    ):
        raise SPSDKError(
            "error: Signed image requires private key with -s option, "
            "one or more certificate(s) using -S option and one or more root key "
            "certificates using -R option"
        )

    # Versions and build number are up to the user. If he doesn't provide any,
    # we set these to following values.
    product_version = parsed_bd_file["options"].get("productVersion", "")
    component_version = parsed_bd_file["options"].get("componentVersion", "")
    build_number = parsed_bd_file["options"].get("buildNumber", -1)

    if not product_version:
        product_version = "1.0.0"
        logger.warning("Production version not defined, defaults to '1.0.0'")

    if not component_version:
        component_version = "1.0.0"
        logger.warning("Component version not defined, defaults to '1.0.0'")

    if build_number == -1:
        build_number = 1
        logger.warning("Build number not defined, defaults to '1.0.0'")

    if key_file_path is None:
        # Legacy elf2sb doesn't report no key provided, but this should
        # be definitely reported to tell the user, what kind of key is being
        # used
        logger.warning("No KEK key provided, using a zero KEK key")
        sb_kek = bytes.fromhex("0" * 64)
    else:
        sb_kek = get_key(key_file_path, expected_size=32)

    # validate keyblobs and perform appropriate actions
    keyblobs = parsed_bd_file.get("keyblobs", [])

    # Based on content of parsed BD file, create a BootSectionV2 and assign
    # commands to them.
    # The content of section looks like this:
    # sections: [
    #   {
    #       section_id: <number>,
    #       options: {}, this is left empty for now...
    #       commands: [
    #           {<cmd1>: {<param1>: value, ...}},
    #           {<cmd2>: {<param1>: value, ...}},
    #           ...
    #       ]
    #   },
    #   {
    #       section_id: <number>,
    #       ...
    #   }
    # ]
    sb_sections = []
    bd_sections = parsed_bd_file["sections"]
    for bd_section in bd_sections:
        section_id = bd_section["section_id"]
        commands = []
        for cmd in bd_section["commands"]:
            for key, value in cmd.items():
                # we use a helper function, based on the key ('load', 'erase'
                # etc.) to create a command object. The helper function knows
                # how to handle the parameters of each command.
                cmd_fce = elf2sb_helper21.get_command(key)
                if key in ("keywrap", "encrypt"):
                    keyblob = {"keyblobs": keyblobs}
                    value.update(keyblob)
                cmd = cmd_fce(value)
                commands.append(cmd)

        sb_sections.append(BootSectionV2(section_id, *commands))

    # We have a list of sections and their respective commands, lets create
    # a boot image v2.1 object
    secure_binary = BootImageV21(
        sb_kek,
        *sb_sections,
        product_version=product_version,
        component_version=component_version,
        build_number=build_number,
        flags=flags,
    )

    # create certificate block
    cert_block = CertBlockV2(build_number=build_number)
    for cert_path in signing_certificate_file_paths:
        cert_data = load_certificate_as_bytes(str(cert_path))
        cert_block.add_certificate(cert_data)
    for cert_idx, cert_path in enumerate(root_key_certificate_paths):
        cert_data = load_certificate_as_bytes(str(cert_path))
        cert_block.set_root_key_hash(cert_idx, Certificate(cert_data))

    # We have our secure binary, now we attach to it the certificate block and
    # the private key content
    secure_binary.cert_block = cert_block
    secure_binary.signature_provider = signature_provider

    if hoh_out_path is None:
        hoh_out_path = os.path.join(os.getcwd(), "hash.bin")

    write_file(secure_binary.cert_block.rkht, hoh_out_path, mode="wb")

    return secure_binary.export()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause


"""Master Boot Image."""

import datetime
import logging
import os
from copy import deepcopy
from inspect import isclass
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

from Crypto.Cipher import AES

from spsdk.crypto.signature_provider import SignatureProvider
from spsdk.exceptions import SPSDKParsingError, SPSDKValueError
from spsdk.image import IMG_DATA_FOLDER, MBIMG_SCH_FILE, mbi_mixin
from spsdk.image.exceptions import SPSDKUnsupportedImageType
from spsdk.image.keystore import KeySourceType, KeyStore
from spsdk.image.trustzone import TrustZone, TrustZoneType
from spsdk.utils.crypto.cert_blocks import CertBlockV2, CertBlockV31
from spsdk.utils.misc import align_block, get_key_by_val, load_configuration, write_file
from spsdk.utils.schema_validator import CommentedConfig, ValidationSchemas, check_config

logger = logging.getLogger(__name__)

PLAIN_IMAGE = (0x00, "Plain Image (either XIP or Load-to-RAM)")
SIGNED_RAM_IMAGE = (0x01, "Plain Signed Load-to-RAM Image")
CRC_RAM_IMAGE = (0x02, "Plain CRC Load-to-RAM Image")
ENCRYPTED_RAM_IMAGE = (0x03, "Encrypted Load-to-RAM Image")
SIGNED_XIP_IMAGE = (0x04, "Plain Signed XIP Image")
CRC_XIP_IMAGE = (0x05, "Plain CRC XIP Image")
SIGNED_XIP_NXP_IMAGE = (0x08, "Plain Signed XIP Image NXP Keys")

DEVICE_FILE = os.path.join(IMG_DATA_FOLDER, "database.yaml")

DEBUG_TRACE_ENABLE = False


# pylint: disable=too-many-ancestors
def get_mbi_class(config: Dict[str, Any]) -> Type["MasterBootImage"]:
    """Get Master Boot Image class.

    :raises SPSDKUnsupportedImageType: The invalid configuration.
    :return: MBI Class.
    """
    schema_cfg = ValidationSchemas.get_schema_file(MBIMG_SCH_FILE)
    device_cfg = load_configuration(DEVICE_FILE)
    # Validate needed configuration to recognize MBI class
    check_config(config, [schema_cfg["image_type"], schema_cfg["family"]])
    try:
        target = get_key_by_val(
            config["outputImageExecutionTarget"], device_cfg["map_tables"]["targets"]
        )
        authentication = get_key_by_val(
            config["outputImageAuthenticationType"], device_cfg["map_tables"]["authentication"]
        )
        family = config["family"]

        cls_name = device_cfg["devices"][family]["images"][target][authentication]
    except (KeyError, SPSDKValueError) as exc:
        raise SPSDKUnsupportedImageType(
            "The type of requested Master boot image is not supported for that device."
        ) from exc

    return globals()[cls_name]


def get_mbi_classes(family: str) -> Dict[str, Tuple[Type["MasterBootImage"], str, str]]:
    """Get all Master Boot Image supported classes for chip family.

    :param family: Chip family.
    :raises SPSDKValueError: The invalid family.
    :return: Dictionary with key like image name and values are Tuple with it's MBI Class
        and target and authentication type.
    """
    device_cfg = load_configuration(DEVICE_FILE)
    if not family in device_cfg["devices"]:
        raise SPSDKValueError("Not supported family for Master Boot Image")

    ret: Dict[str, Tuple[Type["MasterBootImage"], str, str]] = {}

    images: Dict[str, Dict[str, str]] = device_cfg["devices"][family]["images"]

    for target in images.keys():
        for authentication in images[target]:
            cls_name = images[target][authentication]

            ret[f"{family}_{target}_{authentication}"] = (
                globals()[cls_name],
                device_cfg["map_tables"]["targets"][target][0],
                device_cfg["map_tables"]["authentication"][authentication][0],
            )

    return ret


def get_all_mbi_classes() -> List[Type["MasterBootImage"]]:
    """Get all Master Boot Image supported classes.

    :return: List with all MBI Classes.
    """
    ret: Set[Type["MasterBootImage"]] = set()

    for var in globals():
        obj = globals()[var]
        if isclass(obj) and issubclass(obj, MasterBootImage) and obj is not MasterBootImage:
            ret.add(obj)

    return sorted(ret, key=lambda x: x.__name__)


def mbi_generate_config_templates(family: str) -> Dict[str, str]:
    """Generate all possible configuration for selected family.

    :param family: Family description.
    :raises SPSDKError: [description]
    :return: Dictionary of individual templates (key is name of template, value is template itself).
    """
    ret: Dict[str, str] = {}
    # 1: Generate all configuration for MBI
    try:
        mbi_classes = get_mbi_classes(family)
    except SPSDKValueError:
        return ret

    for key, mbi in mbi_classes.items():
        mbi_cls, target, authentication = mbi
        schemas = mbi_cls.get_validation_schemas()

        override = {}
        override["family"] = family
        override["outputImageExecutionTarget"] = target
        override["outputImageAuthenticationType"] = authentication
        yaml_data = CommentedConfig(
            f"Master Boot Image Configuration template for {family}, {mbi_cls.IMAGE_TYPE[1]}.",
            schemas,
            override,
        ).export_to_yaml()

        ret[key] = yaml_data

    return ret


def mbi_get_supported_families() -> List[str]:
    """Get supported families by MBI.

    :return: List of supported family names.
    """
    device_cfg = load_configuration(DEVICE_FILE)
    devices: Dict[str, Any] = device_cfg["devices"]
    return list(devices.keys())


class MasterBootImage:
    """Master Boot Image Interface."""

    IMAGE_TYPE = PLAIN_IMAGE

    app: Optional[bytes]
    app_table: Optional[mbi_mixin.MultipleImageTable]
    collect_data: Callable[[], bytes]
    encrypt: Any  # encrypt(self, raw_image: bytes, revert: bool = False) -> bytes
    post_encrypt: Any  # post_encrypt(self, image: bytes, revert: bool = False) -> bytes
    sign: Any  # sign(self, image: bytes, revert: bool = False) -> bytes
    finalize: Any  # finalize(self, image: bytes, revert: bool = False) -> bytes
    disassemble_image: Callable[[bytes], None]

    @classmethod
    def _get_mixins(cls) -> List[Type[mbi_mixin.Mbi_Mixin]]:
        """Get the list of Mbi Mixin classes.

        :return: List of Mbi_Mixins.
        """
        return [x for x in cls.__bases__ if issubclass(x, mbi_mixin.Mbi_Mixin)]

    def __init__(self) -> None:
        """Initialization of MBI."""
        # Check if all needed class instance members are available (validation of class due to mixin problems)
        self.search_paths: Optional[List[str]] = None
        self.family = "Unknown"
        self.dek: Optional[str] = None
        for base in self._get_mixins():
            for member in base.NEEDED_MEMBERS:
                assert hasattr(self, member)

    @property
    def total_len(self) -> int:
        """Compute Master Boot Image data length.

        :return: Final image data length.
        """
        ret = 0
        for base in self._get_mixins():
            ret += base.mix_len(self)  # type: ignore
        return ret

    @property
    def app_len(self) -> int:
        """Compute application data length.

        :return: Final image data length.
        """
        ret = 0
        for base in self._get_mixins():
            mix_app_len = base.mix_app_len(self)  # type: ignore
            if mix_app_len < 0:
                mix_app_len = base.mix_len(self)  # type: ignore
            ret += mix_app_len
        return ret

    def load_from_config(
        self, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> None:
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        self.search_paths = search_paths
        self.family = config.get("family", "Unknown")
        for base in self._get_mixins():
            base.mix_load_from_config(self, config)  # type: ignore

    def export(self) -> bytes:
        """Export final bootable image.

        :return: Bootable Image in bytes.
        """
        # 1: Validate the input data
        self.validate()
        # 2: Collect all input data into raw image
        raw_image = self.collect_data()
        if DEBUG_TRACE_ENABLE:
            write_file(raw_image, "export_1_collect.bin", mode="wb")
        # 3: Optionally encrypt the image
        encrypted_image = self.encrypt(raw_image)
        if DEBUG_TRACE_ENABLE:
            write_file(encrypted_image, "export_2_encrypt.bin", mode="wb")
        # 4: Optionally do some post encrypt image updates
        encrypted_image = self.post_encrypt(encrypted_image)
        if DEBUG_TRACE_ENABLE:
            write_file(encrypted_image, "export_3_post_encrypt.bin", mode="wb")
        # 5: Optionally sign image
        signed_image = self.sign(encrypted_image)
        if DEBUG_TRACE_ENABLE:
            write_file(signed_image, "export_4_signed.bin", mode="wb")
        # 6: Finalize image
        final_image = self.finalize(signed_image)
        if DEBUG_TRACE_ENABLE:
            write_file(final_image, "export_5_finalized.bin", mode="wb")

        return final_image

    @staticmethod
    def parse(family: str, data: bytes, dek: Optional[str] = None) -> "MasterBootImage":
        """Parse the final image to individual fields.

        :param family: Device family
        :param data: Final Image in bytes
        :param dek: The decryption key for encrypted images
        :raises SPSDKParsingError: Cannot determinate the decoding class
        :return: MBI parsed class
        """
        # 1: Get the right class to parse MBI
        mbi_classes = get_mbi_classes(family)
        image_type = mbi_mixin.Mbi_MixinIvt.get_image_type(data=data)
        authentication = None
        target = None
        mbi_cls_type = None
        for cls_info in mbi_classes.values():
            if cls_info[0].IMAGE_TYPE[0] == image_type:
                mbi_cls_type = cls_info[0]
                target = cls_info[1]
                authentication = cls_info[2]
                logger.info(
                    "Detected MBI image:\n"
                    f"  Authentication:    {authentication}\n"
                    f"  Target:            {target}"
                )
                break

        if mbi_cls_type == None:
            raise SPSDKParsingError("Unsupported MBI type detected.")

        assert mbi_cls_type
        mbi_cls = mbi_cls_type()
        mbi_cls.family = family
        mbi_cls.dek = dek

        # 2: Parse individual mixins what is possible
        # Solve the order - Wait for the mixins that depends on other and run another round
        mixins_src = mbi_cls._get_mixins()
        while mixins_src:
            mixins = mixins_src.copy()
            mixins_src.clear()
            for mixin in mixins:
                logger.debug(f"Parsing: Mixin {mixin.__name__}.")
                for pre_parsed in mixin.PRE_PARSED:
                    if hasattr(mbi_cls, pre_parsed) and getattr(mbi_cls, pre_parsed) == None:
                        logger.debug(
                            f"Parsing: Mixin {mixin.__name__} has to wait to parse {pre_parsed} mixin."
                        )
                        mixins_src.append(mixin)
                        continue
                mixin.mix_parse(mbi_cls, data)  # type: ignore

        # 3: Revert finalize operation of image
        image = mbi_cls.finalize(data, revert=True)
        if DEBUG_TRACE_ENABLE:
            write_file(image, "parse_1_revert_finalize.bin", mode="wb")
        # 4: Revert optional sign of image
        unsigned_image = mbi_cls.sign(image, revert=True)
        if DEBUG_TRACE_ENABLE:
            write_file(unsigned_image, "parse_2_revert_sign.bin", mode="wb")
        # 5: Revert optional some post encrypt image updates
        encrypted_image = mbi_cls.post_encrypt(unsigned_image, revert=True)
        if DEBUG_TRACE_ENABLE:
            write_file(encrypted_image, "parse_3_revert_post_encrypt.bin", mode="wb")
        # 6: Revert optional encryption of the image
        decrypted_image = mbi_cls.encrypt(encrypted_image, revert=True)
        if DEBUG_TRACE_ENABLE:
            write_file(decrypted_image, "parse_4_revert_encrypt.bin", mode="wb")
        # 7: Disassembly rest of image
        mbi_cls.disassemble_image(decrypted_image)

        return mbi_cls

    def create_config(self, output_folder: str) -> None:
        """Create configuration file and its data files from the MBI class.

        :param output_folder: Output folder to store the parsed data
        """
        cfg_values: Dict[str, Union[str, int]] = {}
        for mixin in self._get_mixins():
            cfg_values.update(mixin.mix_get_config(self, output_folder))  # type: ignore
        mbi_classes = get_mbi_classes(self.family)
        for mbi_class in mbi_classes.values():
            if mbi_class[0].__name__ == self.__class__.__name__:
                target = mbi_class[1]
                authentication = mbi_class[2]
                break

        assert target and authentication

        val_schemas = self.get_validation_schemas()
        cfg_values["family"] = self.family
        cfg_values["outputImageExecutionTarget"] = target
        cfg_values["outputImageAuthenticationType"] = authentication

        yaml_data = CommentedConfig(
            main_title=(
                f"Master Boot Image ({self.__class__.__name__}) recreated configuration from :"
                f"{datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S')}."
            ),
            schemas=val_schemas,
            values=cfg_values,
            export_template=False,
        ).export_to_yaml()

        write_file(yaml_data, os.path.join(output_folder, "mbi_config.yaml"))

    @classmethod
    def get_supported_families(cls) -> List[str]:
        """Create the list of supported families by this class.

        :return: List of supported families.
        """
        ret = set()
        device_cfg = load_configuration(DEVICE_FILE)
        devices: Dict[str, Dict] = device_cfg["devices"]
        for device, dev_val in devices.items():
            images: Dict[str, Dict[str, str]] = dev_val["images"]
            for image in images.values():
                for klass in image.values():
                    if klass == cls.__name__:
                        ret.add(device)

        return list(ret)

    @classmethod
    def get_validation_schemas(cls) -> List[Dict[str, Any]]:
        """Create the validation schema for current image type.

        :return: Validation schema.
        """
        schemas = []
        schema_cfg = ValidationSchemas.get_schema_file(MBIMG_SCH_FILE)
        family_schema = deepcopy(schema_cfg["family"])
        family_schema["properties"]["family"]["enum"] = cls.get_supported_families()
        schemas.append(family_schema)
        schemas.append(deepcopy(schema_cfg["image_type"]))
        schemas.append(deepcopy(schema_cfg["output_file"]))
        for base in cls._get_mixins():
            for sch in base.VALIDATION_SCHEMAS:
                schemas.append(deepcopy(schema_cfg[sch]))
            schemas.extend(deepcopy(base.mix_get_extra_validation_schemas()))

        return schemas

    def validate(self) -> None:
        """Validate the setting of image."""
        for base in self._get_mixins():
            base.mix_validate(self)  # type: ignore


########################################################################################################################
# Master Boot Image Class (LPC55)
########################################################################################################################


# pylint: disable=invalid-name
# pylint: disable=abstract-method
class Mbi_PlainXip(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
):
    """Master Boot Plain XiP Image for LPC55xxx family."""

    def __init__(self, app: Optional[bytes] = None, trust_zone: Optional[TrustZone] = None) -> None:
        """Constructor for Master Boot Plain XiP Image for LPC55xxx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        super().__init__()


class Mbi_CrcXip(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC XiP Image for LPC55xxx family."""

    IMAGE_TYPE = CRC_XIP_IMAGE

    def __init__(self, app: Optional[bytes] = None, trust_zone: Optional[TrustZone] = None) -> None:
        """Constructor for Master Boot CRC XiP Image for LPC55xxx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        super().__init__()


class Mbi_CrcRam(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC RAM Image for LPC55xxx family."""

    IMAGE_TYPE = CRC_RAM_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: Optional[int] = None,
    ) -> None:
        """Constructor for Master Boot CRC XiP Image for LPC55xxx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        super().__init__()


class Mbi_SignedXip(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_MixinCertBlockV2,
    mbi_mixin.Mbi_ExportMixinAppTrustZoneCertBlock,
    mbi_mixin.Mbi_ExportMixinRsaSign,
):
    """Master Boot Signed XiP Image for LPC55xxx family."""

    IMAGE_TYPE = SIGNED_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        cert_block: Optional[CertBlockV2] = None,
        signature_provider: Optional[SignatureProvider] = None,
    ) -> None:
        """Constructor for Master Boot Signed XiP Image for LPC55xxx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param cert_block: Certification block of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.cert_block = cert_block
        self.signature_provider = signature_provider
        super().__init__()


class Mbi_SignedRam(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinCertBlockV2,
    mbi_mixin.Mbi_ExportMixinAppTrustZoneCertBlock,
    mbi_mixin.Mbi_ExportMixinRsaSign,
):
    """Master Boot Signed RAM Image for LPC55xxx family."""

    IMAGE_TYPE = SIGNED_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: Optional[int] = None,
        cert_block: Optional[CertBlockV2] = None,
        signature_provider: Optional[SignatureProvider] = None,
    ) -> None:
        """Constructor for Master Boot Signed XiP Image for LPC55xxx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param cert_block: Certification block of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        self.cert_block = cert_block
        self.signature_provider = signature_provider
        super().__init__()


########################################################################################################################
# Master Boot Image Class (i.MXRT5xx/i.MXRT6xx)
########################################################################################################################


class Mbi_PlainRamRtxxx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinHwKey,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
):
    """Master Boot Plain Image for RTxxx."""

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: Optional[int] = None,
        hwk: bool = False,
    ) -> None:
        """Constructor for Master Boot Plain XiP Image for RTxxx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param hwk: Enable HW user mode keys, defaults to false
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        self.user_hw_key_enabled = hwk
        super().__init__()


class Mbi_PlainSignedRamRtxxx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinRelocTable,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_MixinCertBlockV2,
    mbi_mixin.Mbi_MixinHmacMandatory,
    mbi_mixin.Mbi_MixinKeyStore,
    mbi_mixin.Mbi_MixinHwKey,
    mbi_mixin.Mbi_ExportMixinAppTrustZoneCertBlock,
    mbi_mixin.Mbi_ExportMixinRsaSign,
    mbi_mixin.Mbi_ExportMixinHmacKeyStoreFinalize,
):
    """Master Boot Plain Signed RAM Image for RTxxx family."""

    IMAGE_TYPE = SIGNED_RAM_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        app_table: Optional[mbi_mixin.MultipleImageTable] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: Optional[int] = None,
        cert_block: Optional[CertBlockV2] = None,
        signature_provider: Optional[SignatureProvider] = None,
        hmac_key: Optional[Union[bytes, str]] = None,
        key_store: Optional[KeyStore] = None,
        hwk: bool = False,
    ) -> None:
        """Constructor for Master Boot Plain Signed RAM Image for RTxxx family.

        :param app: Application image data, defaults to None
        :param app_table: Application table for additional application binaries, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param cert_block: Certification block of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        :param hmac_key: HMAC key of image, defaults to None
        :param key_store: Optional KeyStore object for image, defaults to None
        :param hwk: Enable HW user mode keys, defaults to false
        """
        self.app = align_block(app) if app else None
        self.app_table = app_table
        self.load_address = load_addr
        self.tz = trust_zone or TrustZone.enabled()
        self.cert_block = cert_block
        self.signature_provider = signature_provider
        self.hmac_key = bytes.fromhex(hmac_key) if isinstance(hmac_key, str) else hmac_key
        self.key_store = key_store
        self.user_hw_key_enabled = hwk
        super().__init__()


class Mbi_CrcRamRtxxx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinRelocTable,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_MixinHwKey,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC RAM Image for RTxxx family."""

    IMAGE_TYPE = CRC_RAM_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        app_table: Optional[mbi_mixin.MultipleImageTable] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: Optional[int] = None,
        hwk: bool = False,
    ) -> None:
        """Constructor for Master Boot CRC RAM Image for RTxxx family.

        :param app: Application image data, defaults to None
        :param app_table: Application table for additional application binaries, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param hwk: Enable HW user mode keys, defaults to false
        """
        self.app = align_block(app) if app else None
        self.app_table = app_table
        self.tz = trust_zone or TrustZone.enabled()
        self.user_hw_key_enabled = hwk
        self.load_address = load_addr
        super().__init__()


class Mbi_EncryptedRamRtxxx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinRelocTable,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_MixinCertBlockV2,
    mbi_mixin.Mbi_MixinHwKey,
    mbi_mixin.Mbi_MixinKeyStore,
    mbi_mixin.Mbi_MixinHmacMandatory,
    mbi_mixin.Mbi_MixinCtrInitVector,
    mbi_mixin.Mbi_ExportMixinRsaSign,
    mbi_mixin.Mbi_ExportMixinHmacKeyStoreFinalize,
):
    """Master Boot Encrypted RAM Image for RTxxx family."""

    IMAGE_TYPE = ENCRYPTED_RAM_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        app_table: Optional[mbi_mixin.MultipleImageTable] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: Optional[int] = None,
        cert_block: Optional[CertBlockV2] = None,
        signature_provider: Optional[SignatureProvider] = None,
        hmac_key: Optional[Union[bytes, str]] = None,
        key_store: Optional[KeyStore] = None,
        ctr_init_vector: Optional[bytes] = None,
        hwk: bool = False,
    ) -> None:
        """Constructor for Master Boot Encrypted RAM Image for RTxxx family..

        :param app: Application image data, defaults to None
        :param app_table: Application table for additional application binaries, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param cert_block: Certification block of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        :param hwk: Enable HW user mode keys, defaults to false
        :param key_store: Optional KeyStore object for image, defaults to None
        :param hmac_key: HMAC key of image, defaults to None
        :param ctr_init_vector: Counter initialization vector of image, defaults to None
        """
        self.app = align_block(app) if app else None
        self.load_address = load_addr
        self.app_table = app_table
        self.tz = trust_zone or TrustZone.enabled()
        self.cert_block = cert_block
        self.signature_provider = signature_provider
        self.user_hw_key_enabled = hwk
        self.key_store = key_store
        self.hmac_key = bytes.fromhex(hmac_key) if isinstance(hmac_key, str) else hmac_key
        self.store_ctr_init_vector(ctr_init_vector)
        self.img_len = 0
        super().__init__()

    def collect_data(self) -> bytes:
        """Collect basic data to create image.

        :return: Collected raw image.
        """
        assert self.cert_block
        self.cert_block.alignment = 4  # type: ignore

        self.img_len = (
            self.total_len + self.cert_block.signature_size + 56 + 16
        )  # Encrypted IVT + IV

        image = self.update_ivt(
            app_data=self.get_app_data() + self.tz.export(),
            total_len=self.img_len,
            crc_val_cert_offset=self.get_app_length(),
        )
        return image

    def disassemble_image(self, image: bytes) -> None:  # pylint: disable=no-self-use
        """Disassemble image to individual parts from image.

        :param image: Image.
        """
        # Re -parse decrypted TZ if needed
        if self.tz.type == TrustZoneType.CUSTOM:
            self.tz = TrustZone.from_binary(
                family=self.family, raw_data=image[-TrustZone.get_preset_data_size(self.family) :]
            )

        tz_len = len(self.tz.export())
        if tz_len:
            image = image[:-tz_len]
        if hasattr(self, "disassembly_app_data"):
            image = self.disassembly_app_data(image)
        self.app = self.clean_ivt(image)

    def encrypt(self, image: bytes, revert: bool = False) -> bytes:
        """Encrypt image if needed.

        :param image: Input raw image to encrypt.
        :param revert: Revert the operation if possible.
        :return: Encrypted image.
        """
        if revert and not (self.hmac_key and self.ctr_init_vector):
            logger.warning("Cannot parse the encrypted image without decrypting key!")
            return image

        assert self.hmac_key and self.ctr_init_vector
        key = self.hmac_key
        if not self.key_store or self.key_store.key_source == KeySourceType.OTP:
            key = KeyStore.derive_enc_image_key(key)
        aes = AES.new(key, AES.MODE_CTR, initial_value=self.ctr_init_vector, nonce=bytes())
        logger.debug(f"Encryption key: {self.hmac_key.hex()}")
        logger.debug(f"Encryption IV: {self.ctr_init_vector.hex()}")

        if revert:
            decrypted_data = aes.decrypt(image)
            return decrypted_data

        encrypted_data = aes.encrypt(image + self.tz.export())
        return encrypted_data

    def post_encrypt(self, image: bytes, revert: bool = False) -> bytes:
        """Optionally do some post encrypt image updates.

        :param image: Encrypted image.
        :param revert: Revert the operation if possible.
        :return: Updated encrypted image.
        """
        assert self.cert_block and isinstance(self.cert_block, CertBlockV2)
        if revert:
            cert_blk_offset = mbi_mixin.Mbi_MixinIvt.get_cert_block_offset(image)
            cert_blk_size = self.cert_block.expected_size
            # Restore original part of encrypted IVT
            org_image = image[
                cert_blk_offset + cert_blk_size : cert_blk_offset + cert_blk_size + 56
            ]
            # Add rest of original encrypted image
            org_image += image[56:cert_blk_offset]
            # optionally add TrustZone data
            org_image += image[cert_blk_offset + cert_blk_size + 56 + 16 :]
            return org_image

        enc_ivt = self.update_ivt(
            app_data=image[: self.HMAC_OFFSET],
            total_len=self.img_len,
            crc_val_cert_offset=self.get_app_length(),
        )

        # Create encrypted cert block (Encrypted IVT table + IV)
        encrypted_header = image[:56] + self.ctr_init_vector

        self.cert_block.image_length = (
            len(image) + len(self.cert_block.export()) + len(encrypted_header)
        )
        enc_cert = self.cert_block.export() + encrypted_header
        return (
            enc_ivt
            + image[self.HMAC_OFFSET : self.get_app_length()]  # header  # encrypted image
            + enc_cert  # certificate + encoded image header + CTR init vector
            + image[self.get_app_length() :]  # TZ encoded data
        )


class Mbi_PlainXipRtxxx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_MixinHwKey,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
):
    """Master Boot Plain XiP Image for RTxxx."""

    def __init__(
        self, app: Optional[bytes] = None, trust_zone: Optional[TrustZone] = None, hwk: bool = False
    ) -> None:
        """Constructor for Master Boot Plain XiP Image for RTxxx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param hwk: Enable HW user mode keys, defaults to false
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.user_hw_key_enabled = hwk
        super().__init__()


class Mbi_PlainSignedXipRtxxx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_MixinCertBlockV2,
    mbi_mixin.Mbi_MixinHwKey,
    mbi_mixin.Mbi_ExportMixinAppTrustZoneCertBlock,
    mbi_mixin.Mbi_ExportMixinRsaSign,
):
    """Master Boot Plain Signed XiP Image for RTxxx family."""

    IMAGE_TYPE = SIGNED_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: Optional[int] = None,
        cert_block: Optional[CertBlockV2] = None,
        signature_provider: Optional[SignatureProvider] = None,
        hwk: bool = False,
    ) -> None:
        """Constructor for Master Boot Plain Signed XiP Image for RTxxx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param cert_block: Certification block of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        :param hwk: Enable HW user mode keys, defaults to false
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        self.cert_block = cert_block
        self.signature_provider = signature_provider
        self.user_hw_key_enabled = hwk
        super().__init__()


class Mbi_CrcXipRtxxx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinTrustZone,
    mbi_mixin.Mbi_MixinHwKey,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC XiP Image for RTxxx."""

    IMAGE_TYPE = CRC_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: Optional[int] = None,
        hwk: bool = False,
    ) -> None:
        """Constructor for Master Boot CRC XiP Image for RTxxx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param hwk: Enable HW user mode keys, defaults to false
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        self.user_hw_key_enabled = hwk
        super().__init__()


########################################################################################################################
# Master Boot Image Class (LPC55x3x)
########################################################################################################################
class Mbi_PlainRamLpc55s3x(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
):
    """Master Boot Plain RAM Image for LPC55s3x family."""

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: int = 0,
        firmware_version: int = 0,
    ) -> None:
        """Constructor for Master Boot Plain RAM Image for LPC55s3x family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.app_ext_memory_align = 0x1000
        self.tz = trust_zone or TrustZone.enabled()
        self.firmware_version = firmware_version
        self.load_address = load_addr
        super().__init__()


class Mbi_CrcRamLpc55s3x(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC RAM Image for LPC55s3x family."""

    IMAGE_TYPE = CRC_RAM_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: int = 0,
        firmware_version: int = 0,
    ) -> None:
        """Constructor for Master Boot Signed RAM Image for LPC55s3x family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.app_ext_memory_align = 0x1000
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        self.firmware_version = firmware_version
        super().__init__()


class Mbi_PlainXipSignedLpc55s3x(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinCertBlockV31,
    mbi_mixin.Mbi_MixinManifest,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppCertBlockManifest,
    mbi_mixin.Mbi_ExportMixinEccSign,
):
    """Master Boot Signed XIP Image for LPC55s3x family."""

    IMAGE_TYPE = SIGNED_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        firmware_version: int = 0,
        load_addr: int = 0,
        cert_block: Optional[CertBlockV31] = None,
        manifest: Optional[mbi_mixin.MasterBootImageManifest] = None,
        signature_provider: Optional[SignatureProvider] = None,
    ) -> None:
        """Constructor for Master Boot Signed XIP Image for LPC55s3x family.

        :param app: Application image data, defaults to None
        :param firmware_version: Firmware version of image, defaults to 0
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param cert_block: Certification block of image, defaults to None
        :param manifest: Manifest of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        """
        self.app = align_block(app) if app else None
        self.app_ext_memory_align = 0x1000
        self.firmware_version = firmware_version
        self.load_address = load_addr
        self.cert_block = cert_block
        self.manifest = manifest
        self.signature_provider = signature_provider
        super().__init__()


class Mbi_CrcXipLpc55s3x(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC XiP Image for LPC55s3x family."""

    IMAGE_TYPE = CRC_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: int = 0,
        firmware_version: int = 0,
    ) -> None:
        """Constructor for Master Boot Signed RAM Image for LPC55s3x family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.app_ext_memory_align = 0x1000
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        self.firmware_version = firmware_version
        super().__init__()


########################################################################################################################
# Master Boot Image Class (KW45xx/K32W1xx)
########################################################################################################################
class Mbi_PlainXipKw45xx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinImageSubType,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
):
    """Master Boot Plain RAM Image for KW45xx/K32W1xx family."""

    def __init__(
        self,
        app: Optional[bytes] = None,
        load_addr: Optional[int] = None,
        trust_zone: Optional[TrustZone] = None,
        firmware_version: int = 0,
        image_subtype: Optional[mbi_mixin.Mbi_MixinImageSubType.Mbi_ImageSubTypeKw45xx] = None,
    ) -> None:
        """Constructor for Master Boot Plain RAM Image for KW45xx/K32W1xx family.

        :param app: Application image data, defaults to None
        :param load_addr: Load/Execution address of image, defaults to 0
        :param trust_zone: TrustZone object, defaults to None
        :param firmware_version: Firmware version of image, defaults to 0
        :param image_subtype: Selection of image subtype (MAIN/NBU), default to None(MAIN)
        """
        self.app = align_block(app) if app else None
        self.load_address = load_addr
        self.tz = trust_zone or TrustZone.enabled()
        self.firmware_version = firmware_version
        self.set_image_subtype(image_subtype)
        super().__init__()


class Mbi_PlainXipSignedKw45xx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinCertBlockV31,
    mbi_mixin.Mbi_MixinManifest,
    mbi_mixin.Mbi_MixinNoSignature,
    mbi_mixin.Mbi_ExportMixinAppCertBlockManifest,
    mbi_mixin.Mbi_ExportMixinEccSign,
):
    """Master Boot Signed XIP Image for KW45xx/K32W1xx family."""

    IMAGE_TYPE = SIGNED_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        load_addr: Optional[int] = None,
        firmware_version: int = 0,
        cert_block: Optional[CertBlockV31] = None,
        manifest: Optional[mbi_mixin.MasterBootImageManifest] = None,
        signature_provider: Optional[SignatureProvider] = None,
        no_signature: Optional[bool] = None,
    ) -> None:
        """Constructor for Master Boot Signed XIP Image for KW45xx/K32W1xx family.

        :param app: Application image data, defaults to None
        :param load_addr: Load/Execution address of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        :param cert_block: Certification block of image, defaults to None
        :param manifest: Manifest of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        :param nxp_image: If it set, the image is generated as NXP.
        :param no_signature: If it set, the signature is not appended
        """
        self.app = align_block(app) if app else None
        self.load_address = load_addr
        self.no_signature = no_signature
        self.firmware_version = firmware_version
        self.cert_block = cert_block
        self.manifest = manifest
        self.signature_provider = signature_provider
        super().__init__()


class Mbi_PlainXipSignedNxpKw45xx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinCertBlockV31,
    mbi_mixin.Mbi_MixinManifest,
    mbi_mixin.Mbi_MixinImageSubType,
    mbi_mixin.Mbi_MixinNoSignature,
    mbi_mixin.Mbi_ExportMixinAppCertBlockManifest,
    mbi_mixin.Mbi_ExportMixinEccSign,
):
    """Master Boot Signed XIP Image by NXP Keys for KW45xx/K32W1xx family."""

    IMAGE_TYPE = SIGNED_XIP_NXP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        load_addr: Optional[int] = None,
        firmware_version: int = 0,
        cert_block: Optional[CertBlockV31] = None,
        manifest: Optional[mbi_mixin.MasterBootImageManifest] = None,
        signature_provider: Optional[SignatureProvider] = None,
        image_subtype: Optional[mbi_mixin.Mbi_MixinImageSubType.Mbi_ImageSubTypeKw45xx] = None,
        no_signature: Optional[bool] = None,
    ) -> None:
        """Constructor for Master Boot Signed XIP Image for KW45xx/K32W1xx family.

        :param app: Application image data, defaults to None
        :param load_addr: Load/Execution address of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        :param cert_block: Certification block of image, defaults to None
        :param manifest: Manifest of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        :param image_subtype: Selection of image subtype (MAIN/NBU), default to None(MAIN)
        :param nxp_image: If it set, the image is generated as NXP.
        :param no_signature: If it set, the signature is not appended
        """
        self.app = align_block(app) if app else None
        self.load_address = load_addr
        self.no_signature = no_signature
        self.firmware_version = firmware_version
        self.cert_block = cert_block
        self.manifest = manifest
        self.signature_provider = signature_provider
        self.set_image_subtype(image_subtype)
        super().__init__()


class Mbi_CrcXipKw45xx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinImageSubType,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC XiP Image for KW45xx/K32W1xx family."""

    IMAGE_TYPE = CRC_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        load_addr: Optional[int] = None,
        trust_zone: Optional[TrustZone] = None,
        firmware_version: int = 0,
        image_subtype: Optional[mbi_mixin.Mbi_MixinImageSubType.Mbi_ImageSubTypeKw45xx] = None,
    ) -> None:
        """Constructor for Master Boot CRC XiP Image for KW45xx/K32W1xx family.

        :param app: Application image data, defaults to None
        :param load_addr: Load/Execution address of image, defaults to 0
        :param trust_zone: TrustZone object, defaults to None
        :param firmware_version: Firmware version of image, defaults to 0
        :param image_subtype: Selection of image subtype (MAIN/NBU), default to None(MAIN)
        """
        self.app = align_block(app) if app else None
        self.load_address = load_addr
        self.tz = trust_zone or TrustZone.enabled()
        self.firmware_version = firmware_version
        self.set_image_subtype(image_subtype)
        super().__init__()


########################################################################################################################
# Master Boot Image Class (MCXNX)
########################################################################################################################
class Mbi_CrcXipMcxNx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC XiP Image for mcxnx family."""

    IMAGE_TYPE = CRC_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        firmware_version: int = 0,
        load_addr: Optional[int] = None,
    ) -> None:
        """Constructor for Master Boot CRC XiP Image for mcxnx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.firmware_version = firmware_version
        self.load_address = load_addr
        self.app_ext_memory_align = 0x1000
        super().__init__()


class Mbi_PlainXipSignedMcxNx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinCertBlockV31,
    mbi_mixin.Mbi_MixinManifestMcxNx,
    mbi_mixin.Mbi_MixinImageSubType,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppCertBlockManifest,
    mbi_mixin.Mbi_ExportMixinEccSign,
):
    """Master Boot Signed XIP Image for mcxnx family."""

    IMAGE_TYPE = SIGNED_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        firmware_version: int = 0,
        cert_block: Optional[CertBlockV31] = None,
        manifest: Optional[mbi_mixin.MasterBootImageManifestMcxNx] = None,
        signature_provider: Optional[SignatureProvider] = None,
        image_subtype: Optional[mbi_mixin.Mbi_MixinImageSubType.Mbi_ImageSubTypeMcxn9xx] = None,
        load_addr: Optional[int] = None,
    ) -> None:
        """Constructor for Master Boot Signed XIP Image for mcxnx family.

        :param app: Application image data, defaults to None
        :param firmware_version: Firmware version of image, defaults to 0
        :param cert_block: Certification block of image, defaults to None
        :param manifest: Manifest of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        :param image_subtype: Selection of image subtype (MAIN/RECOVERY), default to None(MAIN)
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.firmware_version = firmware_version
        self.cert_block = cert_block
        self.manifest = manifest
        self.signature_provider = signature_provider
        self.set_image_subtype(image_subtype)
        self.load_address = load_addr
        self.app_ext_memory_align = 0x1000
        super().__init__()


class Mbi_PlainRamMcxNx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
):
    """Master Boot Plain RAM Image for mcxnx family."""

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: int = 0,
        firmware_version: int = 0,
    ) -> None:
        """Constructor for Master Boot Plain RAM Image for mcxnx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        self.firmware_version = firmware_version
        self.app_ext_memory_align = 0x1000
        super().__init__()


class Mbi_CrcRamMcxNx(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC RAM Image for mcxnx family."""

    IMAGE_TYPE = CRC_RAM_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: int = 0,
        firmware_version: int = 0,
    ) -> None:
        """Constructor for Master Boot Signed RAM Image for mcxnx family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        self.firmware_version = firmware_version
        super().__init__()


########################################################################################################################
# Master Boot Image Class (RW61x)
########################################################################################################################
class Mbi_PlainRamRw61x(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
):
    """Master Boot Plain RAM Image for RW61x family."""

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: int = 0,
        firmware_version: int = 0,
    ) -> None:
        """Constructor for Master Boot Plain RAM Image for RW61x family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.firmware_version = firmware_version
        self.load_address = load_addr
        super().__init__()


class Mbi_CrcRamRw61x(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC RAM Image for RW61x family."""

    IMAGE_TYPE = CRC_RAM_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: int = 0,
        firmware_version: int = 0,
    ) -> None:
        """Constructor for Master Boot Signed RAM Image for RW61x family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        self.firmware_version = firmware_version
        super().__init__()


class Mbi_PlainSignedRamRw61x(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinCertBlockV31,
    mbi_mixin.Mbi_MixinManifest,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppCertBlockManifest,
    mbi_mixin.Mbi_ExportMixinEccSign,
):
    """Master Boot Plain Signed RAM Image for RW61x family."""

    IMAGE_TYPE = SIGNED_RAM_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        load_addr: int = 0,
        firmware_version: int = 0,
        trust_zone: Optional[TrustZone] = None,
        cert_block: Optional[CertBlockV31] = None,
        manifest: Optional[mbi_mixin.MasterBootImageManifest] = None,
        signature_provider: Optional[SignatureProvider] = None,
    ) -> None:
        """Constructor for Master Boot Plain Signed RAM Image for RTxxx family.

        :param app: Application image data, defaults to None
        :param load_addr: Load/Execution address in RAM of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        :param trust_zone: TrustZone object, defaults to None
        :param cert_block: Certification block of image, defaults to None
        :param manifest: Manifest of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        """
        self.app = align_block(app) if app else None
        self.app_ext_memory_align = 0x1000

        self.load_address = load_addr
        self.firmware_version = firmware_version
        self.tz = trust_zone or TrustZone.enabled()
        self.cert_block = cert_block
        self.manifest = manifest
        self.signature_provider = signature_provider
        super().__init__()


class Mbi_PlainExtXipSignedRw61x(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinCertBlockV31,
    mbi_mixin.Mbi_MixinManifest,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppCertBlockManifest,
    mbi_mixin.Mbi_ExportMixinEccSign,
):
    """Master Boot Signed XIP Image stored in external memory for RW61x family."""

    IMAGE_TYPE = SIGNED_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        load_addr: int = 0,
        firmware_version: int = 0,
        trust_zone: Optional[TrustZone] = None,
        cert_block: Optional[CertBlockV31] = None,
        manifest: Optional[mbi_mixin.MasterBootImageManifest] = None,
        signature_provider: Optional[SignatureProvider] = None,
    ) -> None:
        """Constructor for Master Boot Signed XIP Image for RW61x family.

        :param app: Application image data, defaults to None
        :param load_addr: Load/Execution address of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        :param trust_zone: TrustZone object, defaults to None
        :param cert_block: Certification block of image, defaults to None
        :param manifest: Manifest of image, defaults to None
        :param signature_provider: Signature provider to sign final image, defaults to None
        """
        self.app = align_block(app) if app else None
        self.app_ext_memory_align = 0x1000
        self.load_address = load_addr
        self.firmware_version = firmware_version
        self.tz = trust_zone or TrustZone.enabled()
        self.cert_block = cert_block
        self.manifest = manifest
        self.signature_provider = signature_provider
        super().__init__()


class Mbi_CrcExtXipRw61x(
    MasterBootImage,
    mbi_mixin.Mbi_MixinApp,
    mbi_mixin.Mbi_MixinIvt,
    mbi_mixin.Mbi_MixinTrustZoneMandatory,
    mbi_mixin.Mbi_MixinLoadAddress,
    mbi_mixin.Mbi_MixinFwVersion,
    mbi_mixin.Mbi_ExportMixinAppTrustZone,
    mbi_mixin.Mbi_ExportMixinCrcSign,
):
    """Master Boot CRC XiP Image for RW61x family."""

    IMAGE_TYPE = CRC_XIP_IMAGE

    def __init__(
        self,
        app: Optional[bytes] = None,
        trust_zone: Optional[TrustZone] = None,
        load_addr: int = 0,
        firmware_version: int = 0,
    ) -> None:
        """Constructor for Master Boot CRC XiP Image stored in external memory for RW61x family.

        :param app: Application image data, defaults to None
        :param trust_zone: TrustZone object, defaults to None
        :param load_addr: Load/Execution address of image, defaults to 0
        :param firmware_version: Firmware version of image, defaults to 0
        """
        self.app = align_block(app) if app else None
        self.app_ext_memory_align = 0x1000
        self.tz = trust_zone or TrustZone.enabled()
        self.load_address = load_addr
        self.firmware_version = firmware_version
        super().__init__()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Console script for SDPS module aka SDPSHost."""

import sys

import click

from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import (
    CommandsTreeGroup,
    isp_interfaces,
    spsdk_apps_common_options,
)
from spsdk.apps.utils.utils import WARNING_MSG, catch_spsdk_error, get_interface
from spsdk.sdp import SDPS
from spsdk.sdp.sdps import ROM_INFO


@click.group(name="sdpshost", no_args_is_help=True, cls=CommandsTreeGroup)
@isp_interfaces(uart=True, usb=True, is_sdp=True, json_option=False)
@click.option("-n", "--name", type=click.Choice(list(ROM_INFO.keys())), help="Name of the device")
@spsdk_apps_common_options
@click.pass_context
def main(ctx: click.Context, port: str, usb: str, name: str, log_level: int, timeout: int) -> int:
    """Utility for communication with ROM on i.MX targets using SDPS protocol (i.MX8/9)."""
    spsdk_logger.install(level=log_level)
    click.echo(WARNING_MSG)
    # if --help is provided anywhere on command line, skip interface lookup and display help message
    if "--help " in sys.argv:
        port, usb = None, None  # type: ignore
    ctx.obj = {
        "interface": get_interface(module="sdp", port=port, usb=usb, timeout=timeout)
        if port or usb
        else None,
        "name": name,
    }
    return 0


@main.command()
@click.argument("bin_file", metavar="FILE", type=click.File("rb"), required=True)
@click.pass_context
def write_file(ctx: click.Context, bin_file: click.File) -> None:
    """Write boot image data.

    \b
    FILE    - binary file to write
    """
    data = bin_file.read()  # type: ignore
    with SDPS(ctx.obj["interface"], device_name=ctx.obj["name"]) as sdps:
        sdps.write_file(data)


@catch_spsdk_error
def safe_main() -> None:
    """Call the main function."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for general utilities used by applications."""

import logging
import os
import re
import sys
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import click
import hexdump

from spsdk import SPSDKError
from spsdk.mboot import interfaces as MBootInterfaceModule
from spsdk.mboot.interfaces import MBootInterface
from spsdk.sdp import interfaces as SDPInterfaceModule
from spsdk.sdp.interfaces import SDPInterface
from spsdk.utils.misc import (  # pylint: disable=unused-import #backward-compatibility
    find_file,
    get_abs_path,
    load_binary,
    load_file,
    value_to_bytes,
    write_file,
)

WARNING_MSG = """
!!! THIS IS AN EXPERIMENTAL UTILITY! USE WITH CAUTION !!!
"""

logger = logging.getLogger(__name__)


class SPSDKAppError(SPSDKError):
    """Non-fatal error for applications. Sets CLI application error code to 1."""

    fmt = "{description}"

    def __init__(self, desc: Optional[str] = None, error_code: int = 1) -> None:
        """Initialize the AppError.

        :param desc: Description to print out on command line, defaults to None
        :param error_code: Error code passed to OS, defaults to 1
        """
        super().__init__(desc)
        self.description = desc
        self.error_code = error_code


class INT(click.ParamType):
    """Type that allows integers in bin, hex, oct format including _ as a visual separator."""

    name = "integer"

    def __init__(self, base: int = 0) -> None:
        """Initialize custom INT param class.

        :param base: requested base for the number, defaults to 0
        """
        super().__init__()
        self.base = base

    # pylint: disable=inconsistent-return-statements
    def convert(
        self,
        value: str,
        param: Optional[click.Parameter] = None,
        ctx: Optional[click.Context] = None,
    ) -> int:
        """Perform the conversion str -> int.

        :param value: value to convert
        :param param: Click parameter, defaults to None
        :param ctx: Click context, defaults to None
        :return: value as integer
        :raises TypeError: Value is not a string
        :raises ValueError: Value can't be interpreted as an integer
        """
        try:
            return int(value, self.base)
        except TypeError:
            self.fail(
                "expected string for int() conversion, got "
                f"{value!r} of type {type(value).__name__}",
                param,
                ctx,
            )
        except ValueError:
            self.fail(f"{value!r} is not a valid integer", param, ctx)


def get_sdp_interface(
    port: Optional[str] = None, usb: Optional[str] = None, timeout: int = 5000
) -> SDPInterface:
    """Get SDP interface.

    'port', 'usb' parameters are mutually exclusive; one of them is required.

    :param port: name and speed of the serial port (format: name[,speed]), defaults to None
    :param usb: PID,VID of the USB interface, defaults to None
    :param timeout: timeout in milliseconds
    :return: Selected interface instance
    :raises SPSDKError: Only one of appropriate interfaces must be specified
    :raises SPSDKError: Interface couldn't be opened
    """
    all_interfaces = (port, usb)
    count_interfaces = sum(i is not None for i in all_interfaces)

    if count_interfaces == 0:
        raise SPSDKError("One of '--port', '--usb' must be specified.")
    if count_interfaces > 1:
        raise SPSDKError("Only one of '--port', '--usb' must be specified.")

    devices: Union[List[SDPInterfaceModule.RawHid], List[SDPInterfaceModule.Uart]]
    if port:
        port_parts = port.split(",")
        name = port_parts.pop(0)
        baudrate = int(port_parts.pop(), 0) if port_parts else None

        devices = SDPInterfaceModule.scan_uart(port=name, baudrate=baudrate, timeout=timeout)
        if len(devices) != 1:
            raise SPSDKError(f"Cannot ping device on UART port '{name}'.")
    if usb:
        vid_pid = usb.replace(",", ":")
        devices = SDPInterfaceModule.scan_usb(vid_pid)
        if len(devices) == 0:
            raise SPSDKError(f"Cannot find USB device '{format_vid_pid(vid_pid)}'")
        if len(devices) > 1:
            raise SPSDKError(f"More than one device '{format_vid_pid(vid_pid)}' found")
        devices[0].timeout = timeout

    assert devices  # assert just to satisfy pyright, case if handled at the begging
    return devices[0]


def get_mboot_interface(
    port: Optional[str] = None,
    usb: Optional[str] = None,
    sdio: Optional[str] = None,
    timeout: int = 5000,
    buspal: Optional[str] = None,
    lpcusbsio: Optional[str] = None,
) -> MBootInterface:
    """Get appropriate interface.

    'port', 'usb', 'sdio', 'lpcusbsio' parameters are mutually exclusive; one of them is required.

    :param port: name and speed of the serial port (format: name[,speed]), defaults to None
    :param usb: PID,VID of the USB interface, defaults to None
    :param sdio: SDIO path of the SDIO interface, defaults to None
    :param buspal: buspal interface settings, defaults to None
    :param timeout: timeout in milliseconds
    :param lpcusbsio: LPCUSBSIO spi or i2c config string
    :return: Selected interface instance
    :raises SPSDKError: Only one of the appropriate interfaces must be specified
    :raises SPSDKError: When SPSDK-specific error occurs
    """
    # check that one and only one interface is defined
    all_interfaces = (port, usb, sdio, lpcusbsio)
    count_interfaces = sum(i is not None for i in all_interfaces)

    if count_interfaces == 0:
        raise SPSDKError("One of '--port', '--usb', '--sdio' or '--lpcusbsio' must be specified.")
    if count_interfaces > 1:
        raise SPSDKError(
            "Only one of '--port', '--usb', '--sdio' or '--lpcusbsio' must be specified."
        )

    devices: Union[
        List[MBootInterfaceModule.Uart],
        List[MBootInterfaceModule.RawHid],
        List[MBootInterfaceModule.Sdio],
        List[MBootInterfaceModule.UsbSio],
        List[MBootInterfaceModule.MBootInterface],
    ]
    if port:
        port_parts = port.split(",")
        name = port_parts.pop(0)
        baudrate = int(port_parts.pop(), 0) if port_parts else None

        if buspal:
            props = buspal.split(",")
            target = props.pop(0)
            if target == "i2c":
                devices = MBootInterfaceModule.scan_buspal_i2c(
                    port=name, timeout=timeout, props=props
                )
            elif target == "spi":
                devices = MBootInterfaceModule.scan_buspal_spi(
                    port=name, timeout=timeout, props=props
                )
            else:
                raise SPSDKError(f"Target '{target}' is not supported, should be spi or i2c")

            if len(devices) != 1:
                raise SPSDKError(
                    f"Error: cannot communicate with BUSPAL target on UART port '{name}'."
                )
        else:
            devices = MBootInterfaceModule.scan_uart(port=name, baudrate=baudrate, timeout=timeout)
            if len(devices) != 1:
                raise SPSDKError(f"Cannot ping device on UART port '{name}'.")
    if usb:
        vid_pid = usb.replace(",", ":")
        devices = MBootInterfaceModule.scan_usb(vid_pid)
        if len(devices) == 0:
            raise SPSDKError(f"Cannot find USB device '{format_vid_pid(vid_pid)}'")
        if len(devices) > 1:
            raise SPSDKError(f"More than one device '{format_vid_pid(vid_pid)}' found")
        devices[0].timeout = timeout
    if sdio:
        devices = MBootInterfaceModule.scan_sdio(sdio)
        if len(devices) == 0:
            raise SPSDKError(f"Cannot find SDIO device '{sdio}'")
        if len(devices) > 1:
            raise SPSDKError(f"More than one device '{sdio}' found")
        devices[0].timeout = timeout
    if lpcusbsio:
        devices = MBootInterfaceModule.scan_usbsio(lpcusbsio, timeout=timeout)
        if len(devices) != 1:
            raise SPSDKError(
                f"Cannot initialize USBSIO device '{lpcusbsio}',"
                f" exactly one device has to be specified, found: {devices}. "
            )
    assert devices  # assert just to satisfy pyright, case if handled at the begging
    return devices[0]


def get_interface(
    module: str,
    port: Optional[str] = None,
    usb: Optional[str] = None,
    sdio: Optional[str] = None,
    timeout: int = 5000,
    buspal: Optional[str] = None,
    lpcusbsio: Optional[str] = None,
) -> Union[MBootInterface, SDPInterface]:
    """Get appropriate interface.

    'port', 'usb', 'sdio', 'lpcusbsio' parameters are mutually exclusive; one of them is required.

    :param module: name of module to get interface from, 'sdp' or 'mboot'
    :param port: name and speed of the serial port (format: name[,speed]), defaults to None
    :param usb: PID,VID of the USB interface, defaults to None
    :param sdio: SDIO path of the SDIO interface, defaults to None
    :param buspal: buspal interface settings, defaults to None
    :param timeout: timeout in milliseconds
    :param lpcusbsio: LPCUSBSIO spi or i2c config string
    :return: Selected interface instance
    :raises SPSDKError: Only one of the appropriate interfaces must be specified
    :raises SPSDKError: When SPSDK-specific error occurs
    """
    if module == "mboot":
        return get_mboot_interface(
            port=port, usb=usb, sdio=sdio, timeout=timeout, buspal=buspal, lpcusbsio=lpcusbsio
        )
    if module == "sdp":
        return get_sdp_interface(port=port, usb=usb, timeout=timeout)
    raise SPSDKError(f"Unknown interface module: {module}. Valid options are: mboot, sdp")


def _split_string(string: str, length: int) -> list:
    """Split the string into chunks of same length."""
    return [string[i : i + length] for i in range(0, len(string), length)]


def format_raw_data(data: bytes, use_hexdump: bool = False, line_length: int = 16) -> str:
    """Format bytes data into human-readable form.

    :param data: Data to format
    :param use_hexdump: Use hexdump with addresses and ASCII, defaults to False
    :param line_length: bytes per line, defaults to 32
    :return: formatted string (multilined if necessary)
    """
    if use_hexdump:
        return hexdump.hexdump(data, result="return")
    data_string = data.hex()
    parts = [_split_string(line, 2) for line in _split_string(data_string, line_length * 2)]
    result = "\n".join(" ".join(line) for line in parts)
    return result


def format_vid_pid(dec_version: str) -> str:
    """Format VID:PID information in more human-readable format."""
    if ":" in dec_version:
        vid, pid = dec_version.split(":")
        return f"{int(vid, 0):#06x}:{int(pid, 0):#06x}"
    return dec_version


def catch_spsdk_error(function: Callable) -> Callable:
    """Catch the SPSDKError."""

    @wraps(function)
    def wrapper(*args: tuple, **kwargs: dict) -> Any:
        try:
            retval = function(*args, **kwargs)
            return retval
        except SPSDKAppError as app_exc:
            if app_exc.description:
                click.echo(str(app_exc))
            sys.exit(app_exc.error_code)
        except (AssertionError, SPSDKError) as spsdk_exc:
            click.echo(f"ERROR:{spsdk_exc}", err=True)
            logger.debug(str(spsdk_exc), exc_info=True)
            sys.exit(2)
        except UnicodeEncodeError as encode_exc:
            logger.warning(
                (
                    "Your terminal (Jupyter notebook) doesn't render UTF-8 symbols correctly.\n"
                    "Please add the following environment variable and restart any opened shells.\n"
                    "PYTHONIOENCODING=utf8"
                )
            )
            logger.debug(str(encode_exc), exc_info=True)
            sys.exit(2)
        except (Exception, KeyboardInterrupt) as base_exc:  # pylint: disable=broad-except
            click.echo(f"GENERAL ERROR: {type(base_exc).__name__}: {base_exc}", err=True)
            logger.debug(str(base_exc), exc_info=True)
            sys.exit(3)

    return wrapper


def parse_file_and_size(file_and_size: str) -> Tuple[str, int]:
    """Parse composite file-size params.

    :param file_and_size: original param that possibly contains size constrain
    :return: Tuple of path as str and size as int (if present)
    """
    if "," in file_and_size:
        file_path, size = file_and_size.split(",")
        file_size = int(size, 0)
    else:
        file_path = file_and_size
        file_size = -1
    return file_path, file_size


def parse_hex_data(hex_data: str) -> bytes:
    """Parse hex-data into bytes.

    :param hex_data: input hex-data, e.g: {{1122}}, {{11 22}}
    :raises SPSDKError: Failure to parse given input
    :return: data parsed from input
    """
    hex_data = hex_data.replace(" ", "")
    if not hex_data.startswith(("{{", "[[")) or not hex_data.endswith(("}}", "]]")):
        raise SPSDKError("Incorrectly formatted hex-data: Need to start with {{ and end with }}")

    hex_data = hex_data.replace("{{", "").replace("}}", "").replace("[[", "").replace("]]", "")
    if not re.fullmatch(r"[0-9a-fA-F]*", hex_data):
        raise SPSDKError("Incorrect hex-data: Need to have valid hex string")

    str_parts = [hex_data[i : i + 2] for i in range(0, len(hex_data), 2)]
    byte_pieces = [int(part, 16) for part in str_parts]
    result = bytes(byte_pieces)
    if not result:
        raise SPSDKError("Incorrect hex-data: Unable to get any data")
    return bytes(byte_pieces)


# pylint: disable=inconsistent-return-statements
def check_file_exists(path: str, force_overwrite: bool = False) -> bool:  # type: ignore
    """Check if file exists, exits if file exists and overwriting is disabled.

    :param path: Path to a file
    :param force_overwrite: Allows file overwriting
    :raises SPSDKAppError: File already exists and overwriting is disabled
    :return: if file overwriting is allowed, it return True if file exists
    """
    if force_overwrite:
        return os.path.isfile(path)
    if os.path.isfile(path) and not force_overwrite:
        raise SPSDKAppError(f"File '{path}' already exists. Use --force to overwrite it.")


def get_key(
    key_source: Union[str, int], expected_size: int, search_paths: Optional[List[str]] = None
) -> bytes:
    """Get the key from the command line parameter.

    :param key_source: File path to key file or hexadecimal value. If not specified random value is used.
    :param expected_size: Expected size of key in bytes.
    :param search_paths: List of paths where to search for the file, defaults to None
    :raises SPSDKError: Invalid key
    :return: Key in bytes.
    """
    key = None
    assert expected_size > 0, "Invalid expected size of key"
    if isinstance(key_source, int):
        return value_to_bytes(key_source, byte_cnt=expected_size)

    if not key_source:
        logger.debug(
            f"The key source is not specified, the random value is used in size of {expected_size} B."
        )
        # pylint: disable=import-outside-toplevel
        from spsdk.utils.crypto.common import crypto_backend

        return crypto_backend().random_bytes(expected_size)
    try:
        file_path = find_file(key_source, search_paths=search_paths)
        try:
            str_key = load_file(file_path)
            assert isinstance(str_key, str)
            if not str_key.startswith(("0x", "0X")):
                str_key = "0x" + str_key
            key = value_to_bytes(str_key, byte_cnt=expected_size)
            if len(key) != expected_size:
                raise SPSDKError("Invalid Key size.")
        except (SPSDKError, UnicodeDecodeError):
            key = load_binary(file_path)
    except Exception:
        try:
            if not key_source.startswith(("0x", "0X")):
                key_source = "0x" + key_source
            key = value_to_bytes(key_source, byte_cnt=expected_size)
        except SPSDKError:
            pass

    if key is None or len(key) != expected_size:
        raise SPSDKError(f"Invalid key input: {key_source}")

    return key


def store_key(file_name: str, key: bytes, reverse: bool = False) -> None:
    """Store the key in text hexadecimal and binary format.

    :param file_name: Base file name for the key file. The name will be enriched by *.txt and *.bin extension.
    :param key: The key that should be stored.
    :param reverse: Reverse bytes in binary file
    """
    write_file(key.hex(), file_name + ".txt", mode="w")
    if reverse:  # reverse binary order
        key = bytearray(key)
        key.reverse()
        key = bytes(key)
    write_file(key, file_name + ".bin", mode="wb")


def filepath_from_config(
    config: Dict,
    key: str,
    default_value: str,
    base_dir: str,
    output_folder: str = "",
    file_extension: str = "bin",
) -> str:
    """Get file path from configuration dictionary and append .bin if the value is not blank.

    Function returns the output_folder + filename if the filename does not contain path.
    In case filename contains path, return filename and append ".bin".
    The empty string "" indicates that the user doesn't want the output.
    :param config: Configuration dictionary
    :param key: Name of the key
    :param default_value: default value in case key value is not present
    :param base_dir: base directory for path expansion
    :param output_folder: Output folder, if blank file path from config will be used
    :param file_extension: File extension that will be appended
    :return: filename with appended ".bin" or blank filename ""
    """
    filename = config.get(key, default_value)
    file_extension = "." + file_extension
    if filename == "":
        return filename
    if not os.path.dirname(filename):
        filename = os.path.join(output_folder, filename)
    if not filename.endswith(file_extension):
        filename += file_extension
    return get_abs_path(filename, base_dir)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2016-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for serial communication with a target device using MBoot protocol."""

import logging
import struct
import time
from contextlib import contextmanager
from typing import Generator, List, NamedTuple, Optional, Tuple, Union

from crcmod.predefined import mkPredefinedCrcFun
from serial import Serial, SerialTimeoutException
from serial.tools.list_ports import comports

from spsdk.mboot.commands import CmdPacket, CmdResponse, parse_cmd_response
from spsdk.mboot.exceptions import McuBootConnectionError, McuBootDataAbortError
from spsdk.utils.easy_enum import Enum
from spsdk.utils.exceptions import SPSDKTimeoutError
from spsdk.utils.misc import Timeout

from .base import MBootInterface

logger = logging.getLogger(__name__)


def scan_uart(
    port: Optional[str] = None, baudrate: Optional[int] = None, timeout: Optional[int] = None
) -> List["Uart"]:
    """Scan connected serial ports.

    Returns list of serial ports with devices that respond to PING command.
    If 'port' is specified, only that serial port is checked
    If no devices are found, return an empty list.

    :param port: name of preferred serial port, defaults to None
    :param baudrate: speed of the UART interface, defaults to 56700
    :param timeout: timeout in milliseconds, defaults to 5000
    :return: list of interfaces responding to the PING command
    """
    baudrate = baudrate or 57600
    timeout = timeout or 5000
    if port:
        interface = _check_port(port, baudrate, timeout)
        return [interface] if interface else []
    all_ports = [
        _check_port(comport.device, baudrate, timeout) for comport in comports(include_links=True)
    ]
    return list(filter(None, all_ports))


def _check_port(port: str, baudrate: int, timeout: int) -> Optional["Uart"]:
    """Check if device on comport 'port' responds to PING command.

    :param port: name of port to check
    :param baudrate: speed of the UART interface, defaults to 56700
    :param timeout: timeout in milliseconds
    :return: None if device doesn't respond to PING, instance of Interface if it does
    """
    try:
        logger.debug(f"Checking port: {port}, baudrate: {baudrate}, timeout: {timeout}")
        interface = Uart(port=port, baudrate=baudrate, timeout=timeout)
        interface.open()
        interface.close()
        return interface
    except Exception as e:  # pylint: disable=broad-except
        logger.debug(f"{type(e).__name__}: {e}")
        return None


def calc_crc(data: bytes) -> int:
    """Calculate CRC from the data.

    :param data: data to calculate CRC from
    :return: calculated CRC
    """
    crc_function = mkPredefinedCrcFun("xmodem")
    return crc_function(data)


def to_int(data: bytes, little_endian: bool = True) -> int:
    """Convert bytes into single integer.

    :param data: bytes to convert
    :param little_endian: indicate byte ordering in data, defaults to True
    :return: integer
    """
    byte_order = "little" if little_endian else "big"
    return int.from_bytes(data, byteorder=byte_order)  # type: ignore[arg-type]


class PingResponse(NamedTuple):
    """Special type of response for Ping Command."""

    version: int
    options: int
    crc: int

    @classmethod
    def parse(cls, data: bytes) -> "PingResponse":
        """Parse raw data into PingResponse object.

        :param data: bytes to be unpacked to PingResponse object
            4B version, 2B data, 2B CRC16
        :raises McuBootConnectionError: Received invalid ping response
        :return: PingResponse
        """
        try:
            version, options, crc = struct.unpack("<I2H", data)
        except struct.error as err:
            raise McuBootConnectionError("Received invalid ping response") from err
        return cls(version, options, crc)


PING_TIMEOUT_MS = 500
MAX_PING_RESPONSE_DUMMY_BYTES = 50
MAX_UART_OPEN_ATTEMPTS = 3


########################################################################################################################
# UART Interface Class
########################################################################################################################
class FPType(Enum):
    """Type of frames used in serial communication."""

    ACK = 0xA1
    NACK = 0xA2
    ABORT = 0xA3
    CMD = 0xA4
    DATA = 0xA5
    PING = 0xA6
    PINGR = 0xA7


class Uart(MBootInterface):
    """UART interface."""

    FRAME_START_BYTE = 0x5A
    FRAME_START_BYTE_NOT_READY = 0x00

    @property
    def is_opened(self) -> bool:
        """Return True if device is open, False otherwise."""
        return self.device.is_open

    def __init__(
        self, port: Optional[str] = None, baudrate: int = 57600, timeout: int = 5000
    ) -> None:
        """Initialize the UART interface.

        :param port: name of the serial port, defaults to None
        :param baudrate: baudrate of the serial port, defaults to 57600
        :param timeout: read/write timeout in milliseconds, defaults to 5000
        :raises McuBootConnectionError: when the port could not be opened
        """
        super().__init__()
        try:
            self.timeout = timeout
            timeout_s = timeout / 1000
            self.device = Serial(
                port=port, timeout=timeout_s, write_timeout=timeout_s, baudrate=baudrate
            )
            if port:
                self.close()  # TODO: [SPSDK-1212] Is this really necessary here? Please advice somebody.
            self.protocol_version = 0
            self.options = 0
        except Exception as e:
            raise McuBootConnectionError(str(e)) from e

    def open(self) -> None:
        """Open the UART interface.

        :raises McuBootConnectionError: In any case of fail of UART open operation.
        """
        for i in range(MAX_UART_OPEN_ATTEMPTS):
            try:
                self.device.open()
                self.ping()
                logger.debug(f"Interface opened after {i + 1} attempts.")
                return
            except TimeoutError as e:
                self.device.reset_input_buffer()
                self.device.reset_output_buffer()
                # Closing may take up 30-40 seconds
                self.device.close()
                logger.debug(f"Timeout when pinging the device: {repr(e)}")
            except McuBootConnectionError as e:
                self.device.close()
                logger.debug(f"Opening interface failed with: {repr(e)}")
            except Exception as exc:
                self.device.close()
                raise McuBootConnectionError("UART Interface open operation fails.") from exc
        raise McuBootConnectionError(
            f"Cannot open UART interface after {MAX_UART_OPEN_ATTEMPTS} attempts."
        )

    def close(self) -> None:
        """Close the UART interface.

        :raises McuBootConnectionError: In any case of fail of UART close operation.
        """
        try:
            self.device.close()
        except Exception as e:
            raise McuBootConnectionError(str(e)) from e

    def info(self) -> str:
        """Return information about the UART interface.

        :return: Description of UART interface
        :raises McuBootConnectionError: When no port is available
        """
        try:
            return self.device.port
        except Exception as e:
            raise McuBootConnectionError(str(e)) from e

    def read(self) -> Union[CmdResponse, bytes]:
        """Read data from device.

        :return: read data
        :raises McuBootDataAbortError: Indicates data transmission abort
        :raises McuBootConnectionError: When received invalid CRC
        """
        _, frame_type = self._read_frame_header()
        length = to_int(self._read_default(2))
        crc = to_int(self._read_default(2))
        if not length:
            self._send_ack()
            raise McuBootDataAbortError()
        data = self._read_default(length)
        self._send_ack()
        calculated_crc = self._calc_frame_crc(data, frame_type)
        if crc != calculated_crc:
            raise McuBootConnectionError("Received invalid CRC")
        if frame_type == FPType.CMD:
            return parse_cmd_response(data)
        return data

    def write(self, packet: Union[CmdPacket, bytes]) -> None:
        """Write data to the device; data might be in form of 'CmdPacket' or bytes.

        :param packet: Packet to send
        :raises AttributeError: frame type is incorrect
        """
        data, frame_type = None, None
        if isinstance(packet, CmdPacket):
            data = packet.to_bytes(padding=False)
            frame_type = FPType.CMD
        if isinstance(packet, (bytes, bytearray)):
            data = packet
            frame_type = FPType.DATA
        if not data or not frame_type:
            raise AttributeError("Incorrect packet type")
        frame = self._create_frame(data, frame_type)
        self._send_frame(frame, wait_for_ack=True)

    def _read_default(self, length: int) -> bytes:
        """Read 'length' amount of bytes from device, this function can be overridden in child class.

        :param length: Number of bytes to read
        :type length: int
        :return: Data read from the device
        """
        return self._read(length)

    def _read(self, length: int) -> bytes:
        """Read 'length' amount for bytes from device.

        :param length: Number of bytes to read
        :return: Data read from the device
        :raises SPSDKTimeoutError: Time-out
        :raises McuBootConnectionError: When reading data from device fails
        """
        try:
            data = self.device.read(length)
        except Exception as e:
            raise McuBootConnectionError(str(e)) from e
        if not data:
            raise SPSDKTimeoutError()
        logger.debug(f"<{' '.join(f'{b:02x}' for b in data)}>")
        return data

    def _write(self, data: bytes) -> None:
        """Send data to device.

        :param data: Data to send
        :raises SPSDKTimeoutError: Time-out
        :raises McuBootConnectionError: When sending the data fails
        """
        logger.debug(f"[{' '.join(f'{b:02x}' for b in data)}]")
        try:
            self.device.reset_input_buffer()
            self.device.reset_output_buffer()
            self.device.write(data)
            self.device.flush()
        except SerialTimeoutException as e:
            raise SPSDKTimeoutError(
                f"Write timeout error. The timeout is set to {self.device.write_timeout} s. Consider increasing it."
            ) from e
        except Exception as e:
            raise McuBootConnectionError(str(e)) from e

    def _send_ack(self) -> None:
        ack_frame = struct.pack("<BB", self.FRAME_START_BYTE, FPType.ACK)
        self._send_frame(ack_frame, wait_for_ack=False)

    def _send_frame(self, frame: bytes, wait_for_ack: bool = True) -> None:
        """Send a frame to UART.

        :param frame: Data to send
        :param wait_for_ack: Wait for ACK frame from device, defaults to True
        """
        self._write(frame)
        if wait_for_ack:
            self._read_frame_header(FPType.ACK)

    def _read_frame_header(self, expected_frame_type: Optional[int] = None) -> Tuple[int, int]:
        """Read frame header and frame type. Return them as tuple of integers.

        :param expected_frame_type: Check if the frame_type is exactly as expected
        :return: Tuple of integers representing frame header and frame type
        :raises McuBootDataAbortError: Target sens Data Abort frame
        :raises McuBootConnectionError: Unexpected frame header or frame type (if specified)
        :raises McuBootConnectionError: When received invalid ACK
        """
        timeout = Timeout(self.timeout, "ms")
        header = -1
        while not timeout.overflow():
            header = to_int(self._read_default(1))
            if header != self.FRAME_START_BYTE_NOT_READY:
                break
        # This is workaround addressing SPI ISP issue on RT5/6xx when sometimes
        # ACK frames and START BYTE frames are swapped, see SPSDK-1824 for more details
        if header not in [self.FRAME_START_BYTE, FPType.ACK]:
            raise McuBootConnectionError(
                f"Received invalid frame header '{header:#X}' expected '{self.FRAME_START_BYTE:#X}'"
                + "\nTry increasing the timeout, some operations might take longer"
            )
        if header == FPType.ACK:
            frame_type = header
        else:
            frame_type = to_int(self._read(1))
        if frame_type == FPType.ABORT:
            raise McuBootDataAbortError()
        if expected_frame_type:
            if frame_type == self.FRAME_START_BYTE:
                frame_type = header
            if frame_type != expected_frame_type:
                raise McuBootConnectionError(
                    f"received invalid ACK '{frame_type:#X}' expected '{expected_frame_type:#X}'"
                )
        return header, frame_type

    def _create_frame(self, data: bytes, frame_type: int) -> bytes:
        """Encapsulate data into frame."""
        crc = self._calc_frame_crc(data, frame_type)
        frame = struct.pack(
            f"<BBHH{len(data)}B",
            self.FRAME_START_BYTE,
            frame_type,
            len(data),
            crc,
            *data,
        )
        return frame

    def _calc_frame_crc(self, data: bytes, frame_type: int) -> int:
        """Calculate the CRC of a frame.

        :param data: frame data
        :param frame_type: frame type
        :return: calculated CRC
        """
        crc_data = struct.pack(
            f"<BBH{len(data)}B", self.FRAME_START_BYTE, frame_type, len(data), *data
        )
        return calc_crc(crc_data)

    def ping(self) -> None:
        """Ping the target device, retrieve protocol version.

        :raises McuBootConnectionError: If the target device doesn't respond to ping
        :raises McuBootConnectionError: If the start frame is not received
        :raises McuBootConnectionError: If the header is invalid
        :raises McuBootConnectionError: If the frame type is invalid
        :raises McuBootConnectionError: If the ping response is not received
        :raises McuBootConnectionError: If crc does not match
        """
        with self.ping_timeout(timeout=PING_TIMEOUT_MS):
            ping = struct.pack("<BB", self.FRAME_START_BYTE, FPType.PING)
            self._send_frame(ping, wait_for_ack=False)

            # after power cycle, MBoot v 3.0+ may respond to first command with a leading dummy data
            # we read data from UART until the FRAME_START_BYTE byte
            start_byte = b""
            for i in range(MAX_PING_RESPONSE_DUMMY_BYTES):
                start_byte = self._read_default(1)
                if start_byte is None:
                    raise McuBootConnectionError("Failed to receive initial byte")

                if start_byte == self.FRAME_START_BYTE.to_bytes(length=1, byteorder="little"):
                    logger.debug(f"FRAME_START_BYTE received in {i + 1}. attempt.")
                    break
            else:
                raise McuBootConnectionError("Failed to receive FRAME_START_BYTE")

            header = to_int(start_byte)
            if header != self.FRAME_START_BYTE:
                raise McuBootConnectionError("Header is invalid")
            frame_type = to_int(self._read_default(1))
            if frame_type != FPType.PINGR:
                raise McuBootConnectionError("Frame type is invalid")

            response_data = self._read_default(8)
            if response_data is None:
                raise McuBootConnectionError("Failed to receive ping response")
            response = PingResponse.parse(response_data)

            # ping response has different crc computation than the other responses
            # that's why we can't use calc_frame_crc method
            # crc data for ping excludes the last 2B of response data, which holds the CRC from device
            crc_data = struct.pack(
                f"<BB{len(response_data) -2}B", header, frame_type, *response_data[:-2]
            )
            crc = calc_crc(crc_data)
            if crc != response.crc:
                raise McuBootConnectionError("Received CRC doesn't match")

            self.protocol_version = response.version
            self.options = response.options

    @contextmanager
    def ping_timeout(self, timeout: int = PING_TIMEOUT_MS) -> Generator[None, None, None]:
        """Context manager for changing UART's timeout.

        :param timeout: New temporary timeout in milliseconds, defaults to PING_TIMEOUT_MS (500ms)
        :return: Generator[None, None, None]
        """
        context_timeout = min(timeout, self.timeout)
        context_timeout_s = context_timeout / 1000
        self.device.timeout = context_timeout_s
        self.device.write_timeout = context_timeout_s
        logger.debug(f"Setting timeout to {context_timeout} ms")
        # driver needs to be reconfigured after timeout change, wait for a little while
        time.sleep(0.005)

        yield

        restored_timeout_s = self.timeout / 1000
        self.device.timeout = restored_timeout_s
        self.device.write_timeout = restored_timeout_s
        logger.debug(f"Restoring timeout to {self.timeout} ms")
        time.sleep(0.005)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module provides exceptions for PFR module."""
from spsdk import SPSDKError


class SPSDKUnsupportedImageType(SPSDKError):
    """The specified Image type is not supported."""
